""" 
 
CSV data to HEAL VLMD conversion


""" 
from healdata_utils.io import read_delim 
from healdata_utils.transforms.jsontemplate.conversion import convert_templatejson
from healdata_utils.types import typesets
import pandas as pd
def convert_datacsv(file_path,data_dictionary_props={}):
    """ 
    Takes a CSV file containing data (not metadata) and 
    infers each of it's variables data types and names.
    These inferred properties are then outputted as partially-completed HEAL variable level metadata
    files. That is, it outputs the `name` and `type` property. 

    NOTE: this will be an invalid file as `description` is required
    for each variable. However, this serves as a great way to start
    the basis of a VLMD submission.
    """  
    df = read_delim(file_path)
    data_dictionary = data_dictionary_props.copy()
    fields = typesets.infer_frictionless_fields(df)
    data_dictionary['fields'] = fields 

    package = convert_templatejson(data_dictionary)
    return package



""" General utilities/helper functions"""
import re
from collections.abc import MutableMapping
import pandas as pd
from pandas.api.types import is_object_dtype

def _get_propnames_to_rearrange(propnames,schema):
    """ 
    get data dictionary props to search for refactoring or embedding
    
    Supports patternProperties
    """

    annotation_names = ["title","description","name","additionalDescription"]
    root_names = flatten_properties(schema["properties"]).keys()
    field_names = flatten_properties(schema["properties"]["fields"]["items"]["properties"]).keys()
    props_to_rearrange = set(root_names).intersection(field_names).difference(annotation_names)

    propnames_exp = "|".join("^"+name+"$" for name in props_to_rearrange)
    names_to_rearrange = [name for name in propnames if re.match(propnames_exp,name)]

    return list(names_to_rearrange)

def embed_data_dictionary_props(flat_fields,flat_root,schema):
    """ 
    Embed (flattened) root level props in each (flattened) field
    if field level prop is missing but the field level prop exists
    and is not an annotation property (title,description). 
    
    Params
    
    flat_fields: array of dicts or pd.DataFrame (or something that can be converted into DataFrame)
    flat_root: dict or pd.Series or something that can be converted into Series 
    schema: schema for determining what fields should be embedded (note, this is flattened in fxn)
    
    Returns
    
    pd.DataFrame with the flat fields with the embedded root properties
    """
    flat_fields = pd.DataFrame(flat_fields)
    propnames = _get_propnames_to_rearrange(list(flat_root.keys()),schema)
    flat_root = pd.Series(flat_root).loc[propnames] # take out annotation props
    if len(flat_root) > 0:
        for propname in propnames:
            if propname in flat_root:
                if not propname in flat_fields:
                    flat_fields.insert(0,propname,flat_root[propname])
                else:
                    flat_fields[propname].fillna(flat_root[propname],inplace=True)

    return flat_fields

def refactor_field_props(flat_fields,schema):
    """ 
    given a flattened array of dicts corresponding to the unflattened schema,
    move up (ie `refactor`) flattened properties that are both in the root 
    (ie table level; level up from field records) and in the fields.

    """  
    flat_fields_df = pd.DataFrame(flat_fields)
    propnames = set(_get_propnames_to_rearrange(flat_fields_df.columns.tolist(),schema))
    flat_record = pd.Series(dtype="object")
    for name in propnames:
        in_df = name in flat_fields_df
        if in_df:
            # need to handle if some values are pandas series
            if isinstance(flat_fields_df[name], pd.DataFrame):
                is_one_unique = (flat_fields_df[name].nunique() == 1).all()
            elif isinstance(flat_fields_df[name], pd.Series):
                if is_object_dtype(flat_fields_df[name]):
                    is_one_unique = len(flat_fields_df[name].map(str).unique()) == 1
                else:
                    is_one_unique = flat_fields_df[name].nunique() == 1
            if is_one_unique:
                flat_record[name] = flat_fields_df.pop(name).iloc[0]
                if isinstance(flat_record[name], pd.Series):
                    flat_record[name] = flat_record[name].to_list()
            

    return flat_record,flat_fields_df

# individual cell utilities
def strip_html(html_string):
    if html_string:
        return re.sub(r"<[^>]+>", "", html_string)
    else:
        return html_string


def to_int_if_base10(val):
    """
    converts value to a string and if
    float (or a string rep of a float) of base10
    to an integer string representation.

    NOTE:
    """
    string = str(val)

    if "." in string:
        parts = string.split(".")
        if len(parts) == 2 and parts[1] == "0":
            return parts[0]

    return string


def parse_dictionary_str(string, item_sep, keyval_sep):
    """
    parses a stringified dictionary into a dictionary
    based on item separator

    """
    if string != "" and string != None:
        stritems = string.strip().split(item_sep)
        items = {}

        for stritem in stritems:
            if stritem:
                item = stritem.split(keyval_sep, 1)
                items[item[0].strip()] = item[1].strip()
        
        return items
    else:
        return string


def parse_list_str(string, item_sep):
    if string != "" and string != None:
        return string.strip().split(item_sep)
    else:
        return string


# dictionary utilities
def flatten_to_jsonpath(dictionary,schema,parent_key=False, sep="."):
    """
    Turn a nested dictionary into a flattened dictionary (but see schema param)

    :param dictionary: The dictionary to flatten
    :param schema: The schema to indicate which properties to flatten
        This includes dictionaries (properties) with child dictionaries (properties)
        and lists (items) with child dictionaries (properties)
    :param parent_key: The string to prepend to dictionary's keys
    :param sep: The string used to separate flattened keys 
    :return: A flattened dictionary
    """
    # flatten if type array -> type object with properties
    # flatten if type object with properties
    items = []
    for key, value in dictionary.items():
        new_key = str(parent_key) + sep + key if parent_key else key
        prop = schema["properties"].get(key,{})
        childprops = prop.get("properties")
        childitem_props = prop.get("items",{}).get("properties")

        if childitem_props:
            for i,_value in enumerate(value):
                item = flatten_to_jsonpath(
                    dictionary=_value, 
                    schema=prop,
                    parent_key=new_key, 
                    sep=f"[{str(i)}]{sep}")
                items.extend(item.items())
        elif childprops:
            item = flatten_to_jsonpath(
                dictionary=value, 
                schema=prop,
                parent_key=new_key, 
                sep=sep)
            items.extend(item.items())

        else:
            items.append((new_key, value))

    return dict(items)

def stringify_keys(dictionary):
    orig_keys = list(dictionary.keys())
    for key in orig_keys:
        dictionary[str(key)] = dictionary.pop(key)


def unflatten_from_jsonpath(field):
    """
    Converts a flattened dictionary with key names conforming to 
    JSONpath notation to the nested dictionary format.
    """
    field_json = {}

    for prop_path, prop in field.items():
        prop_json = field_json

        nested_names = prop_path.split(".")
        for prop_name in nested_names[:-1]:
            if '[' in prop_name:
                array_name, array_index = prop_name.split('[')
                array_index = int(array_index[:-1])
                if array_name not in prop_json:
                    prop_json[array_name] = [{} for _ in range(array_index + 1)]
                elif len(prop_json[array_name]) <= array_index:
                    prop_json[array_name].extend([{} for _ in range(array_index - len(prop_json[array_name]) + 1)])
                prop_json = prop_json[array_name][array_index]
            else:
                if prop_name not in prop_json:
                    prop_json[prop_name] = {}
                prop_json = prop_json[prop_name]

        last_prop_name = nested_names[-1]
        if '[' in last_prop_name:
            array_name, array_index = last_prop_name.split('[')
            array_index = int(array_index[:-1])
            if array_name not in prop_json:
                prop_json[array_name] = [{} for _ in range(array_index + 1)]
            elif len(prop_json[array_name]) <= array_index:
                prop_json[array_name].extend([{} for _ in range(array_index - len(prop_json[array_name]) + 1)])
            if isinstance(prop_json[array_name][array_index], dict):
                prop_json[array_name][array_index].update({array_name: prop})
            else:
                prop_json[array_name][array_index] = {array_name: prop}
        else:
            prop_json[last_prop_name] = prop

    return field_json
    
# json to csv utils
def join_iter(iterable,sep_list="|"):
    return sep_list.join([str(p) for p in iterable])

def join_dictvals(dictionary:dict,sep:str):
    return sep.join(dictionary.values())

def join_dictitems(dictionary:dict,sep_keyval='=',sep_items='|'):
    """ joins a mappable collection (ie dictionary) into a string
    representation with specified separators for the key and value
    in addition to items. 

    All items are coerced to the string representation (eg if key or value
    is None, this will be coerced to "None")


    """
    dict_list = []
    for key,val in dictionary.items():
        keystr = str(key)
        valstr = str(val)
        dict_list.append(keystr+sep_keyval+valstr)
    return sep_items.join(dict_list)


# documentation building utilities
def find_docstring_desc(fxn):
    """
    return the description part of a docstring
    (ie text before Parameters)
    """
    exp = "^(.*)Parameters\\n"

    if fxn.__doc__:
        docstring = fxn.__doc__.strip()
    else:
        docstring = "No documentation"
    try:
        return re.search(exp, docstring, re.DOTALL).group(1)
    except AttributeError:
        return docstring


# add missing values and order according to the order of a list


def sync_fields(data, field_list,missing_value=None):
    """
    Sorts fields and adds missing fields (with None value).
    If extra fields exist in a record that are not in field_list, then tacks on at 
    end of record.


    Parameters
    --------------
    data [list]: json array of values
    field_list [list]: the list of all field names  (e.g., propertjies from a schema)

    Returns
    -------------
    list: json array with fields added if missing
    """
    data_with_missing = []

    for record in data:
        new_record = {}
        for fieldpropname in field_list:
            # check if the fieldname is a property (or could be a pattern property)
            fieldnames = re.findall("|".join("^"+fieldname+"$" for fieldname in record),fieldpropname)
            # if match then add to new record
            if fieldnames:
                for name in fieldnames:
                    new_record[name] = record[name]
            # if no record than add missing value (if there is a regex list index, then add [0])
            else:
                extra_fieldname = (
                    fieldpropname
                    .replace("^","")
                    .replace("$","")
                    .replace("\[\d+\]","[0]")
                ) #replace list item regex
                new_record[extra_fieldname] = missing_value

        # tack on extra fields not in field_list at back
        extra_fields = list(set(list(record)).difference(list(new_record)))
        for name in extra_fields:
            new_record[name] = record[name]

        # append the newly synced record with missing values
        data_with_missing.append(new_record)

    
    return data_with_missing


# %% 
# Working with schemas
def flatten_properties(properties, parentkey="", sep=".",itemsep="\[\d+\]"):
    """
    flatten schema properties
    """
    properties_flattened = {}
    for key, item in properties.items():
        # flattened keys
        if parentkey:
            flattenedkey = parentkey + "." + key
        else:
            flattenedkey = key

        if isinstance(item, MutableMapping):
            props = item.get("properties")
            items = item.get("items",{}).get("properties")
            if props:
                newprops = flatten_properties(props, parentkey=flattenedkey)
                properties_flattened.update(newprops)

            elif items:
                newprops = flatten_properties(items,parentkey=flattenedkey+itemsep)
                properties_flattened.update(newprops)
            else:
                properties_flattened[flattenedkey] = item
        
        else:
            properties_flattened[flattenedkey] = item
    
    return properties_flattened

def flatten_schema(schema):
    """ 
    Flatten a schema of type object with properties. 
    
    If a regular expression indicating a list index exists, puts this
    under `patternProperties`
    
    """ 
    schema_flattened = dict(schema)
    if "properties" in schema:
        properties = schema_flattened.pop("properties")
        item_sep = "\[\d+\]"
        schema_flattened["properties"] = flatten_properties(properties,itemsep=item_sep)
        schema_flattened["patternProperties"] = {}
        for propname in list(schema_flattened["properties"].keys()):
            if item_sep in propname:
                # TODO: put on table level for csv schema
                # var0 = propname.replace(item_sep,"[0]")
                # var1 = propname.replace(item_sep,"[1]")
                # var2 = propname.replace(item_sep,"[2]")
                # pattern_property_note = (
                #     "\n\n"
                #     "Specifying field names:\n\n"
                #     "This field can have 1 or more columns using the digit index number in brackets (`[0]` --> `[1]` --> `[n]`)\n\n"
                #     "For 1 value, you will have the field (column) names:\n"
                #     "`{0}`\n\n"
                #     # "\tFor 2 values, you will have the columns: "
                #     # "`{0},`{1}`\n"
                #     "For 3 values, you will have the field (column) names:\n"
                #     "`{0}`\t`{1}`\t`{2}`\n\n"
                # ).format(var0,var1,var2)
                pattern_property_note = ""
                pattern_prop = schema_flattened["properties"].pop(propname)
                pattern_prop["description"] = pattern_prop.get("description","") + pattern_property_note
                schema_flattened["patternProperties"]["^"+propname+"$"] = pattern_prop

    return schema_flattened





def find_propname(colname,properties):
    """ 
    given a dictionary of json schema object properties OR a list of property names, return the 
    matching property name. 

    This function is needed when a schema is flattened according to a json path 
    and converted into a regular expression for list (array) indices.

    """ 
    propmatch = []
    for name in list(properties):
        if re.match("^"+name+"$",colname):
            propmatch.append(name)

    if len(propmatch) == 1:
        return propmatch[0]
    elif len(propmatch) > 1:
        raise Exception(f"Multiple matching properties found for {colname}. Can only have one match")
    elif len(propmatch) == 0:
        return None
    else:
        raise Exception(f"Unknown error when matching properties against {colname}")


from ..readstat.conversion import convert_readstat,Path


def convert_spss(file_path,
    data_dictionary_props=None):

    ext = Path(file_path).suffix

    if ext == ".sav":
        return convert_readstat(file_path,data_dictionary_props)
    else:
        raise Exception("Currently, SPSS inputtypes only support SAV files")


healcsvschema = {
    "version": "0.3.2",
    "title": "HEAL Variable Level Metadata Fields",
    "description": "\n"
    "\n"
    '!!! note "Highly encouraged"\n'
    "\n"
    "  - Only `name` and `description` properties are required. \n"
    "  - For categorical variables, `constraints.enum` and "
    "`enumLabels` (where applicable) properties are highly "
    "encouraged. \n"
    "  - For studies using HEAL or other common data elements "
    "(CDEs), `standardsMappings` information is highly "
    "encouraged.\n"
    "  - `type` and `format` properties may be particularly "
    "useful for some variable types (e.g. date-like variables)\n",
    "type": "object",
    "required": ["name", "description"],
    "additionalProperties": False,
    "properties": {
        "schemaVersion": {
            "type": "string",
            "description": "The version of the schema "
            "used in agreed upon "
            "convention of "
            "major.minor.path (e.g., "
            "1.0.2) \n"
            "\n"
            "NOTE: This is NOT for "
            "versioning of each "
            "indiviual data dictionary "
            "instance. \n"
            "Rather, it is the\n"
            "version of THIS schema "
            "document. See `version` "
            "property (below) if "
            "specifying the individual "
            "data dictionary instance\n"
            "version.\n"
            "\n"
            "If generating a vlmd "
            "document as a csv file, "
            "include this version in \n"
            "every row/record to "
            "indicate this is a schema "
            "level property \n"
            "(not applicable for the "
            "json version as this "
            "property is already at "
            "the schema/root level)\n",
            "pattern": "\\d+\\.\\d+\\.\\d+",
            "examples": ["1.0.0", "0.2.0"],
        },
        "section": {
            "type": "string",
            "title": "Section",
            "description": "The section, form, survey "
            "instrument, set of measures  or "
            "other broad category used \n"
            "to group variables. Previously "
            'called "module."\n',
            "examples": ["Demographics", "PROMIS", "Medical History"],
        },
        "name": {
            "type": "string",
            "title": "Variable Name",
            "description": "The name of a variable (i.e., "
            "field) as it appears in the "
            "data. \n",
            "examples": ["gender_id"],
        },
        "title": {
            "type": "string",
            "title": "Variable Label (ie Title)",
            "description": "The human-readable title or label " "of the variable.\n",
            "examples": ["Gender identity"],
        },
        "description": {
            "type": "string",
            "title": "Variable Description",
            "description": "An extended description of "
            "the variable. This could be "
            "the definition of a "
            "variable or the \n"
            "question text (e.g., if a "
            "survey). \n",
            "examples": [
                "The participant's age at the " "time of study enrollment",
                "What is the highest grade or "
                "level of school you have "
                "completed or the highest "
                "degree you have received?",
            ],
        },
        "type": {
            "title": "Variable Type",
            "type": "string",
            "description": "A classification or category of a "
            "particular data element or "
            "property expected or allowed in "
            "the dataset.\n",
            "additionalDescription": "enum definitions:\n"
            "\n"
            "-  `number` (A numeric "
            "value with optional "
            "decimal places. (e.g., "
            "3.14))\n"
            "-  `integer` (A whole "
            "number without decimal "
            "places. (e.g., 42))\n"
            "-  `string` (A sequence "
            "of characters. (e.g., "
            '\\"test\\"))\n'
            "-  `any` (Any type of "
            "data is allowed. (e.g., "
            "true))\n"
            "-  `boolean` (A binary "
            "value representing true "
            "or false. (e.g., true))\n"
            "-  `date` (A specific "
            "calendar date. (e.g., "
            '\\"2023-05-25\\"))\n'
            "-  `datetime` (A "
            "specific date and time, "
            "including timezone "
            "information. (e.g., "
            '\\"2023-05-25T10:30:00Z\\"))\n'
            "-  `time` (A specific "
            "time of day. (e.g., "
            '\\"10:30:00\\"))\n'
            "-  `year` (A specific "
            "year. (e.g., 2023)\n"
            "-  `yearmonth` (A "
            "specific year and month. "
            '(e.g., \\"2023-05\\"))\n'
            "-  `duration` (A length "
            "of time. (e.g., "
            '\\"PT1H\\")\n'
            "-  `geopoint` (A pair of "
            "latitude and longitude "
            "coordinates. (e.g., "
            "[51.5074, -0.1278]))\n",
            "enum": [
                "number",
                "integer",
                "string",
                "any",
                "boolean",
                "date",
                "datetime",
                "time",
                "year",
                "yearmonth",
                "duration",
                "geopoint",
            ],
        },
        "format": {
            "title": "Variable Format",
            "type": "string",
            "description": "Indicates the format of the type "
            "specified in the `type` "
            "property. \n"
            "Each format is dependent on the "
            "`type` specified. \n"
            "See "
            "[here](https://specs.frictionlessdata.io/table-schema/#types-and-formats) \n"
            "for more information about "
            "appropriate `format` values by "
            "variable `type`.\n",
            "additionalDescription": "examples/definitions "
            "of patterns and "
            "possible values:\n"
            "\n"
            "Examples of date time "
            "pattern formats\n"
            "\n"
            "- `%Y-%m-%d` (for "
            "date, e.g., "
            "2023-05-25)\n"
            "- `%Y%-%d` (for date, "
            "e.g., 20230525) for "
            "date without dashes\n"
            "- `%Y-%m-%dT%H:%M:%S` "
            "(for datetime, e.g., "
            "2023-05-25T10:30:45)\n"
            "- `%Y-%m-%dT%H:%M:%SZ` "
            "(for datetime with UTC "
            "timezone, e.g., "
            "2023-05-25T10:30:45Z)\n"
            "- "
            "`%Y-%m-%dT%H:%M:%S%z` "
            "(for datetime with "
            "timezone offset, e.g., "
            "2023-05-25T10:30:45+0300)\n"
            "- `%Y-%m-%dT%H:%M` "
            "(for datetime without "
            "seconds, e.g., "
            "2023-05-25T10:30)\n"
            "- `%Y-%m-%dT%H` (for "
            "datetime without "
            "minutes and seconds, "
            "e.g., 2023-05-25T10)\n"
            "- `%H:%M:%S` (for "
            "time, e.g., 10:30:45)\n"
            "- `%H:%M:%SZ` (for "
            "time with UTC "
            "timezone, e.g., "
            "10:30:45Z)\n"
            "- `%H:%M:%S%z` (for "
            "time with timezone "
            "offset, e.g., "
            "10:30:45+0300)\n"
            "\n"
            "Examples of string "
            "formats\n"
            "\n"
            "- `email` if valid "
            "emails (e.g., "
            "test@gmail.com)\n"
            "- `uri` if valid uri "
            "addresses (e.g., "
            "https://example.com/resource123)\n"
            "- `binary` if a base64 "
            "binary encoded string "
            "(e.g., authentication "
            "token like "
            "aGVsbG8gd29ybGQ=)\n"
            "- `uuid` if a "
            "universal unique "
            "identifier also known "
            "as a guid (eg., "
            "f47ac10b-58cc-4372-a567-0e02b2c3d479)\n"
            "\n"
            "\n"
            "Examples of geopoint "
            "formats\n"
            "\n"
            "The two types of "
            "formats for `geopoint` "
            "(describing a "
            "geographic point).\n"
            "\n"
            "- `array` (if "
            "'lat,long' (e.g., "
            "36.63,-90.20))\n"
            "- `object` (if "
            "{'lat':36.63,'lon':-90.20})\n",
        },
        "constraints.required": {
            "type": "boolean",
            "title": "Required variable",
            "description": "If this variable "
            "is marked as true, "
            "then this "
            "variable's value "
            "must be present\n"
            "(ie not missing; "
            "see "
            "missingValues). If "
            "marked as false or "
            "not present, then "
            "the \n"
            "variable CAN be "
            "missing.\n",
        },
        "constraints.maxLength": {
            "type": "integer",
            "title": "Maximum Length",
            "description": "Indicates the "
            "maximum length of "
            "an iterable "
            "(e.g., array, "
            "string, or\n"
            "object). For "
            "example, if "
            "'Hello World' is "
            "the longest value "
            "of a\n"
            "categorical "
            "variable, this "
            "would be a "
            "maxLength of "
            "11.\n",
        },
        "constraints.enum": {
            "type": "string",
            "title": "Variable Possible Values",
            "description": "Constrains possible " "values to a set of " "values.\n",
            "examples": ["1|2|3|4|5", "Poor|Fair|Good|Very " "good|Excellent"],
            "pattern": "^(?:[^|]+\\||[^|]*)(?:[^|]*\\|)*[^|]*$",
        },
        "constraints.pattern": {
            "type": "string",
            "title": "Regular Expression " "Pattern",
            "description": "A regular "
            "expression pattern "
            "the data MUST "
            "conform to.\n",
        },
        "constraints.maximum": {
            "type": "string",
            "title": "Maximum Value",
            "description": "Specifies the "
            "maximum value of a "
            "field (e.g., "
            "maximum -- or most\n"
            "recent -- date, "
            "maximum integer "
            "etc). Note, this is "
            "different then\n"
            "maxLength "
            "property.\n",
        },
        "constraints.minimum": {
            "type": "string",
            "title": "Minimum Value",
            "description": "Specifies the " "minimum value of a " "field.\n",
        },
        "enumLabels": {
            "title": "Variable Value Encodings (i.e., " "mappings; value labels)",
            "description": "Variable value encodings "
            "provide a way to further "
            "annotate any value within a "
            "any variable type,\n"
            "making values easier to "
            "understand. \n"
            "\n"
            "\n"
            "Many analytic software "
            "programs (e.g., SPSS,Stata, "
            "and SAS) use numerical "
            "encodings and some "
            "algorithms\n"
            "only support numerical "
            "values. Encodings (and "
            "mappings) allow categorical "
            "values to be stored as\n"
            "numerical values.\n"
            "\n"
            "Additionally, as another use "
            "case, this field provides a "
            "way to\n"
            "store categoricals that are "
            'stored as  "short" labels '
            "(such as\n"
            "abbreviations).\n"
            "\n"
            "This field is intended to "
            "follow [this "
            "pattern](https://specs.frictionlessdata.io/patterns/#table-schema-enum-labels-and-ordering)\n",
            "type": "string",
            "examples": [
                "1=Poor|2=Fair|3=Good|4=Very " "good|5=Excellent",
                "HW=Hello world|GBW=Good bye " "world|HM=Hi, Mike",
            ],
            "pattern": "^(?:.*?=.*?(?:\\||$))+$",
        },
        "enumOrdered": {
            "title": "An ordered variable",
            "description": "Indicates whether a "
            "categorical variable is "
            "ordered. This variable  is\n"
            "relevant for variables that "
            "have an ordered "
            "relationship but not\n"
            "necessarily  a numerical "
            "relationship (e.g., "
            "Strongly disagree < "
            "Disagree\n"
            "< Neutral < Agree).\n"
            "\n"
            "This field is intended to "
            "follow the ordering aspect "
            "of this [this pattern][this "
            "pattern](https://specs.frictionlessdata.io/patterns/#table-schema-enum-labels-and-ordering)\n",
            "type": "boolean",
        },
        "missingValues": {
            "title": "Missing Values",
            "description": "A list of missing values " "specific to a variable.\n",
            "examples": ["Missing|Skipped|No " "preference", "Missing"],
            "type": "string",
            "pattern": "^(?:[^|]+\\||[^|]*)(?:[^|]*\\|)*[^|]*$",
        },
        "trueValues": {
            "title": "Boolean True Value Labels",
            "description": "For boolean (true) variable "
            "(as defined in type field), "
            "this field allows\n"
            "a physical string "
            "representation to be cast as "
            "true (increasing\n"
            "readability of the field). "
            "It can include one or more "
            "values.\n",
            "type": "string",
            "examples": ["required|Yes|Checked", "required"],
            "pattern": "^(?:[^|]+\\||[^|]*)(?:[^|]*\\|)*[^|]*$",
        },
        "falseValues": {
            "title": "Boolean False Value Labels",
            "description": "For boolean (false) "
            "variable (as defined in "
            "type field), this field "
            "allows\n"
            "a physical string "
            "representation to be cast "
            "as false (increasing\n"
            "readability of the field) "
            "that is not a standard "
            "false value. It can include "
            "one or more values.\n",
            "type": "string",
            "examples": ["Not required|NOT REQUIRED", "No"],
            "pattern": "^(?:[^|]+\\||[^|]*)(?:[^|]*\\|)*[^|]*$",
        },
        "custom": {
            "type": "string",
            "description": "Additional properties not " "included a core property. \n",
            "pattern": "^(?:.*?=.*?(?:\\||$))+$",
        },
    },
    "patternProperties": {
        "^standardsMappings\\[\\d+\\].instrument.url$": {
            "title": "Url",
            "description": "A "
            "url "
            "(e.g., "
            "link, "
            "address) "
            "to "
            "a "
            "file "
            "or "
            "other "
            "resource "
            "containing "
            "the "
            "instrument, "
            "or\n"
            "a "
            "set "
            "of "
            "items "
            "which "
            "encompass "
            "a "
            "variable "
            "in "
            "this "
            "variable "
            "level "
            "metadata "
            "document "
            "(if "
            "at "
            "the "
            "root "
            "level "
            "or "
            "the "
            "document "
            "level) \n"
            "or "
            "the "
            "individual "
            "variable "
            "(if "
            "at "
            "the "
            "field "
            "level). \n",
            "type": "string",
            "format": "uri",
            "examples": [
                "https://www.heal.nih.gov/files/CDEs/2023-05/adult-demographics-cdes.xlsx"
            ],
        },
        "^standardsMappings\\[\\d+\\].instrument.source$": {
            "type": "string",
            "title": "Source",
            "description": "An "
            "abbreviated "
            "name/acronym "
            "from "
            "a "
            "controlled "
            "vocabulary "
            "referencing "
            "the "
            "resource "
            "(e.g., "
            "program "
            "or "
            "repository)\n"
            "containing "
            "the "
            "instrument, "
            "or "
            "a "
            "set "
            "of "
            "items "
            "which "
            "encompass "
            "a "
            "variable "
            "in "
            "this "
            "variable "
            "level "
            "metadata "
            "document "
            "(if "
            "at "
            "the "
            "root "
            "level "
            "or "
            "the "
            "document "
            "level) \n"
            "or "
            "the "
            "individual "
            "variable "
            "(if "
            "at "
            "the "
            "field "
            "level). \n",
            "enum": ["heal-cde"],
        },
        "^standardsMappings\\[\\d+\\].instrument.title$": {
            "type": "string",
            "title": "Title",
            "examples": ["Adult " "demographics", "adult-demographics"],
            "description": "",
        },
        "^standardsMappings\\[\\d+\\].instrument.id$": {
            "type": "string",
            "title": "Identifier",
            "description": "A "
            "code "
            "or "
            "other "
            "string "
            "that "
            "identifies "
            "the "
            "instrument "
            "within "
            "the "
            "source.\n"
            "This "
            "should "
            "always "
            "be "
            "from "
            "the "
            "source's "
            "formal, "
            "standardized "
            "identification "
            "system \n",
            "examples": ["5141"],
        },
        "^standardsMappings\\[\\d+\\].item.url$": {
            "title": "Standards " "mappings " "- " "Url",
            "description": "The "
            "url "
            "that "
            "links "
            "out "
            "to "
            "the "
            "published, "
            "standardized "
            "mapping "
            "of "
            "a "
            "variable "
            "(e.g., "
            "common "
            "data "
            "element)\n",
            "type": "string",
            "format": "uri",
            "examples": [
                "https://evs.nci.nih.gov/ftp1/CDISC/SDTM/SDTM%20Terminology.html#CL.C74457.RACE"
            ],
        },
        "^standardsMappings\\[\\d+\\].item.source$": {
            "title": "Standards " "mappings " "- " "Source",
            "description": "The "
            "source "
            "of "
            "the "
            "standardized "
            "variable. "
            "Note, "
            "this "
            "property "
            "is "
            "required "
            "if \n"
            "an "
            "id "
            "is "
            "specified.\n",
            "examples": ["CDISC"],
            "type": "string",
        },
        "^standardsMappings\\[\\d+\\].item.id$": {
            "title": "Standards " "Mappings " "- " "Id",
            "type": "string",
            "description": "The "
            "id "
            "locating "
            "the "
            "individual "
            "mapping "
            "within "
            "the "
            "given "
            "source. \n"
            "Note, "
            "the "
            "`standardsMappings[0].source` "
            "property "
            "is "
            "required "
            "if \n"
            "this "
            "property "
            "is "
            "specified.\n",
            "examples": ["C74457"],
        },
        "^relatedConcepts\\[\\d+\\].url$": {
            "title": "Related " "Concepts " "- Url",
            "description": "The "
            "url "
            "that "
            "links "
            "out "
            "to "
            "the "
            "published, "
            "related "
            "concept. \n"
            "The "
            "listed "
            "examples "
            "could "
            "both "
            "be "
            "attached "
            "to "
            "any "
            "variable "
            "related "
            "to, "
            "for "
            "example, "
            "heroin "
            "use.\n",
            "type": "string",
            "format": "uri",
            "examples": [
                "https://www.ebi.ac.uk/chebi/chebiOntology.do?chebiId=CHEBI:27808",
                "http://purl.bioontology.org/ontology/RXNORM/3304",
            ],
        },
        "^relatedConcepts\\[\\d+\\].title$": {
            "title": "Related " "concepts " "- " "Type",
            "description": "A "
            "human-readable "
            "title "
            "(ie "
            "label) "
            "to "
            "a "
            "concept "
            "related "
            "to "
            "the "
            "given "
            "field.\n"
            "The "
            "listed "
            "examples "
            "could "
            "both "
            "be "
            "attached "
            "to "
            "any "
            "variable "
            "related "
            "to, "
            "for "
            "example, "
            "heroin "
            "use.\n",
            "type": "string",
            "examples": ["Heroin " "Molecular " "Structure", "Heroin " "Ontology"],
        },
        "^relatedConcepts\\[\\d+\\].source$": {
            "title": "Related " "Concepts " "- " "Source",
            "description": "The "
            "source "
            "(e.g., "
            "a "
            "dictionary "
            "or "
            "vocabulary "
            "set) "
            "to "
            "a "
            "concept "
            "related "
            "to "
            "the "
            "given "
            "field.\n"
            "The "
            "listed "
            "examples "
            "could "
            "both "
            "be "
            "attached "
            "to "
            "any "
            "variable "
            "related "
            "to, "
            "for "
            "example, "
            "heroin "
            "use.\n",
            "type": "string",
            "examples": ["CHEBI", "RXNORM"],
        },
        "^relatedConcepts\\[\\d+\\].id$": {
            "title": "Related " "Concepts " "- Id",
            "type": "string",
            "description": "The "
            "id "
            "locating "
            "the "
            "individual "
            "concept "
            "within "
            "the "
            "source "
            "of "
            "the "
            "given "
            "field.\n"
            "The "
            "listed "
            "examples "
            "could "
            "both "
            "be "
            "attached "
            "to "
            "any "
            "variable "
            "related "
            "to, "
            "for "
            "example, "
            "heroin "
            "use.\n",
            "examples": ["27808", "3304"],
        },
    },
}




from healdata_utils import mappings 

def test_versions_mapping_is_updated():
    mess = ("Check that you have the most updated schema version for mappings \n",
            f"Current mapping version: {'.'.join(mappings.versions.VERSION)} \n",
            f"Current schema version: {mappings.versions.healjsonschema['version']}")
    assert mappings.versions.VERSION == mappings.versions.healjsonschema["version"].split(".")[0:2],mess


""" utilities to validate data and metadata """ 
import jsonschema
import os
from pathlib import Path
import pandas as pd
import json

from healdata_utils.schemas import healjsonschema,healcsvschema
from healdata_utils.io import read_delim
from healdata_utils.transforms.frictionless.conversion import convert_frictionless_to_jsonschema
from .jsonschema import validate_against_jsonschema
from healdata_utils import utils

class Validator:
    """ 
    input a data object or pointer/path
    to data object with a given schema
    and corresponding type of schema. 
    validate against the said schema of a certain type.
    if the validation schema type is different than the 
    input schema type, will convert to that type (eg jsonschema to
    frictionless.)
    """ 
    @classmethod
    def from_csv_file(cls,path,schema,schema_type):
        data = read_delim(path).to_dict(orient="records")
        return cls(data,schema,schema_type)

    @classmethod
    def from_pandas(cls,data,schema,schema_type):
        data = data.to_dict(orient="records")
        return cls(data,schema,schema_type)
    
    @classmethod 
    def from_jsonarray(cls,data,schema,schema_type):
        jsonschema.validate(data,{"type":"array","items":{"type":"object"}})
        return cls(data,schema,schema_type)

    @classmethod 
    def from_jsonobject(cls,data,schema,schema_type):
        jsonschema.validate(data,{"type":"object"})
        return cls(data,schema,schema_type)
        
    @classmethod
    def from_jsonfile(cls,path,schema,schema_type):
        data = json.loads(Path(path).read_text())
        return cls(data,schema,schema_type)

    def __init__(self,data,schema,schema_type):

        self.data = data #json array
        self.schema = schema
        self.schema_type = schema_type

    def raise_invalid_error(self):
        if not self.report["valid"]:
            raise Exception("These records are not valid")

    def validate(self,against_schema_type):
        if against_schema_type == "jsonschema":
            if self.schema_type == "frictionless":
                schema = convert_frictionless_to_jsonschema(self.schema)
            else:
                schema = self.schema
            report = validate_against_jsonschema(self.data,schema)
        else:
            ## add frictionless 
            ## add pandera -- if needed
            raise NotImplementedError("Still need to implement conversions and validation fxns")

        return {"data":self.data,"schema":schema,"report":report}

def validate_vlmd_json(
    data_or_path,
    schema=healjsonschema,
    input_schema_type="jsonschema",
    validation_schema_type="jsonschema"
    ):
    """
    Validates json data by iterating over every property in every record and comparing to the property 
    specified in the given schema

    Parameters
    ----------
    data_or_path : Path-like object indicating a path to a tabular data source (eg CSV or TSV) or a json array of records (see validate fxn)
    schema : dict, optional
        The schema to compare data_or_path to (default: HEAL frictionless template)

    Returns
    -------
    dict[bool,dict]
        the returned `validate` function object 
    """
    if isinstance(data_or_path, (str, os.PathLike)):
        validator = Validator.from_jsonfile(
            path=data_or_path,
            schema=schema,
            schema_type="jsonschema"
        )
    else:
        validator = Validator.from_jsonobject(
            data=data_or_path,
            schema=schema,
            schema_type="jsonschema"
        )

    package = validator.validate(validation_schema_type)
    report = package["report"]
    # report_summary = []
    errors_list = []
    for error in report["errors"]:
        errors_list.append({"json_path":error["json_path"],
            "message":error["message"]})
        # report_summary.append([error["json_path"],error["message"]])

    package["report"] = {"valid":report["valid"],"errors":errors_list}
    # report_summary_str += "Validation summary for {data_name}"
    # report_summary_str += "\n\n"
    # report_summary_str += "VALID" if report["valid"] else "INVALID"

    # report_summary_str += "## Errors "
    # report_summary_str += "\n\n"

    # report_summary_str+=str(tabulate(
    #     report_summary,
    #     headers=["JsonPath", "Message"],
    #     tablefmt="grid",
    #     maxcolwidths=[5, 5, 90]
    # ))

    # another possible way:
    
    # if not report["valid"]:
    #     print("JSON data dictionary requires modifications:")
    #     console_report = (
    #         pd.DataFrame(data_dictionaries["errors"])
    #         ["jsontemplate"]
    #         ["errors"]
    #         .drop(columns=["json_path"])
    #         .value_counts()
    #     )

    return package
    
def validate_vlmd_csv(
    data_or_path,
    schema={"type":"array","items":healcsvschema},
    to_sync_fields=True
):
    """
    Validates a json array against the heal variable level metadata
    schema catered for a CSV tabular data dictionary file.

    As there are many options to validate a tabular data file, 
    this function provides options for different different specification
    conversions. The default is to input a frictionless schema and run
    validation with jsonschema library 
    NOTE: (the frictionless v4 tools have some
    issues with pyinstaller currently (haven't tried frictionless v5 though)) 

    TODO: replace the csv schema with flattened json json-schema

    Parameters
    ----------
    data_or_path : Path-like object indicating a path to a tabular data source (eg CSV or TSV) or a json array of records (see validate fxn)
    schema : dict, optional
        The schema of type object with all field properties to compare data_or_path to (default: HEAL csv specs).
    input_schema_type: str, optional : the type of schema ["jsonschema","frictionless"]
    validation_schema_type: str, optional : the type of schema to use for validation (will convert if input does not eq validation schema types)
    to_sync_fields : bool, optional[default=True]:whether to add missing fields (ie null) in schema before validation. 
        Note, this is different than missing values. Missing values do not equal missing fields (think tabular dataset)
        This also syncs the order of the fields to order of properties in schema.
    Returns
    -------
    dict[bool,dict]
        the returned `validate` function object 
        (eg., `{"valid":False,"errors":[...]}`)
    """
    def _add_missing_type(propname,prop,schema):
        missing_values = ["",None] # NOTE: include physical rep and logical for now
        if propname in schema.get("required",[]):
            # if required value: MUST be NOT missing value and the property
             newprop = {
                "allOf":[
                    prop,{"not": {"enum":missing_values}}
                ]
            }
        else:
            # if not required value: MUST be property OR the specified missing value
            newprop = {
                "anyOf":[
                    prop,{"enum":missing_values}
            ]}
        return newprop
    # instantiate validator object with correct class method depending on input

    input_schema_type = "jsonschema" # NOTE: prior version was frictionless but changed to jsonschema
    validation_schema_type = "jsonschema"

    props_with_missing = {}
    for propname,prop in schema.get("items",{}).get("properties",{}).items():
        props_with_missing[propname] = _add_missing_type(propname,prop,schema)

    patterns_with_missing = {}
    for patternname,prop in schema.get("items",{}).get("patternProperties",{}).items():
        patterns_with_missing[patternname] = _add_missing_type(patternname,prop,schema)
    
    schema = {"type":"array","items":{}}
    schema["items"]["properties"] = props_with_missing
    schema["items"]["patternProperties"] = patterns_with_missing
            
    if isinstance(data_or_path, (str, os.PathLike)):
        validator = Validator.from_csv_file(path=data_or_path,schema=schema,schema_type=input_schema_type)

    else:
        validator = Validator.from_jsonarray(data=data_or_path,schema=schema,schema_type=input_schema_type)

    # sync fields
    field_list = list(schema["items"].get("properties",[]))
    for name in list(schema["items"].get("patternProperties",[])):
        field_list.append(name)
    
    if to_sync_fields:
        validator.data = utils.sync_fields(validator.data, field_list,missing_value="")

    package = validator.validate(validation_schema_type)
    report = package["report"]
    # report_summary = []
    errors_list = []
    for error in report["errors"]:
        errors_list.append({"json_path":error["json_path"],
            "message":error["message"]})
        # report_summary.append([error["json_path"],error["message"]])

    package["report"] = {"valid":report["valid"],"errors":errors_list}
    # report_summary_str += "Validation summary for {data_name}"
    # report_summary_str += "\n\n"
    # report_summary_str += "VALID" if report["valid"] else "INVALID"

    # report_summary_str += "## Errors "
    # report_summary_str += "\n\n"

    # report_summary_str+=str(tabulate(
    #     report_summary,
    #     headers=["JsonPath", "Message"],
    #     tablefmt="grid",
    #     maxcolwidths=[5, 5, 90]
    # ))

    # another possible way:
    
    # if not report["valid"]:
    #     print("JSON data dictionary requires modifications:")
    #     console_report = (
    #         pd.DataFrame(data_dictionaries["errors"])
    #         ["jsontemplate"]
    #         ["errors"]
    #         .drop(columns=["json_path"])
    #         .value_counts()
    #     )


    return package

    

        





from healdata_utils.io import read_excel,pd
from healdata_utils.types import typesets
from healdata_utils.transforms.jsontemplate.conversion import convert_templatejson
def convert_dataexcel(file_path,data_dictionary_props=None,sheet_name=None,multiple_data_dicts=True):
    """ 
    converts a file or file like object (eg pandas.ExcelFile) into a data dictionary
    package or a dict of data_dictionary packages

    file_path: str - path to xlsx file
    sheet_name: Union[str,list,None]. By default (None), all sheets are read and a dict
        of data frame is returned. If a list of sheets is provided, also returns a dict of packages. Else, returns 
    multiple_data_dicts, boolean: if each sheet represents one data resource 
        (ie if False, all sheets will be concatenated before inference)
    
    """ 

    dfs = read_excel(file_path,sheet_name)

    if isinstance(dfs,pd.DataFrame):
        dfs_to_infer = {"_onlyone":dfs}
        onlyone = True
    else:
        if multiple_data_dicts:
            dfs_to_infer = dfs
            onlyone = False
        else:
            dfs_to_infer = {"_onlyone":pd.concat(dfs.values())}
            onlyone = True
    
    packages = {}

    for name,df in dfs_to_infer.items():
        data_dictionary = {}
        fields = typesets.infer_frictionless_fields(df)

        data_dictionary['fields'] = fields

        package = convert_templatejson(data_dictionary)

    
        packages[name] = package

    if onlyone:
        return packages["_onlyone"]
    else:
        return packages
     





from healdata_utils import utils
import jsonschema


schema_version = {"schemaVersion": {"type": "string"}}
another_field_to_rearrange = {
    "anotherFieldToEmbed": {
        "items": {"type": "object", "properties": {"thisone": {"type": "string"}}}
    }
}
schema_to_rearrange = {
    "version": "0.2.0",
    "$schema": "http://json-schema.org/draft-07/schema#",
    "$id": "test",
    "type": "object",
    "properties": {
        "title": {"type": "string"},
        "description": {"type": "string"},
        **schema_version,
        **another_field_to_rearrange,
        "anotherFieldToEmbed": {
            "items": {
                "type": "object",
                "properties": {"thisone": {"type": "string"}},
            }
        },
        "fields": {
            "type": "array",
            "items": {
                "properties": {
                    "type": "object",
                    "justAField": {"type": "string"},
                    **schema_version,
                    **another_field_to_rearrange,
                }
            },
        },
    },
}


def test_add_missing_fields():
    # data (note one var not in schema)
    data = [
        {"var2": 2, "var3": 3, "var1": 1},
        {
            "var5": 10,
            "var4": 9,
            "var1": 1,
        },
    ]
    # fields from schema
    fields = ["var1", "var2", "var3", "var4"]

    data_with_missing = utils.sync_fields(data, fields)
    assert data_with_missing == [
        {"var1": 1, "var2": 2, "var3": 3, "var4": None},
        {"var1": 1, "var2": None, "var3": None, "var4": 9, "var5": 10},
    ]


def test_unflatten_jsonpath():
    input = {
        "module": "Testing",
        "constraints.enum": "1|2|3|4",
        "standardsMappings[1].item.url": "http//:helloitem1",
        "standardsMappings[0].item.url": "http//:helloitem0",
        "standardsMappings[0].instrument.url": "http//:helloworld0",
        "standardsMappings[1].instrument.url": "http//:helloworld1",
        "standardsMappings[2].item.url": "http//:helloitem2",
        "test1.test2.test3[1]": "test3_1",
        "test1.test2.test3[0].test4": "test4_1",
    }

    output = {
        "module": "Testing",
        "constraints": {"enum": "1|2|3|4"},
        "standardsMappings": [
            {
                "item": {"url": "http//:helloitem0"},
                "instrument": {"url": "http//:helloworld0"},
            },
            {
                "item": {"url": "http//:helloitem1"},
                "instrument": {"url": "http//:helloworld1"},
            },
            {
                "item":{"url":"http//:helloitem2"}
            }
        ],
        "test1": {"test2": {"test3": [{"test4": "test4_1"}, {"test3": "test3_1"}]}},
    }
    field_json = utils.unflatten_from_jsonpath(input)
    assert (
        field_json == output
    ), "Problem with converting input dictionary to output dictionary"


input = {
    "schemaVersion": "0.2.0",
    "anotherFieldToEmbed": [{"thisone": "helloworld"}],
    "fields": [{"justAField": "cool"}, {"justAField": "sad"}],
}


def test_embed_data_dictionary_props():

    flat_root = {
        "schemaVersion": "0.2.0",
        "anotherFieldToEmbed[0].thisone": "helloworld",
    }
    flat_fields_array = [{"justAField": "cool"}, {"justAField": "sad"}]
    flat_fields = utils.embed_data_dictionary_props(
        flat_fields_array, flat_root, schema_to_rearrange
    )

    assert flat_fields.to_dict(orient="records") == [
        {
            "justAField": "cool",
            "schemaVersion": "0.2.0",
            "anotherFieldToEmbed[0].thisone": "helloworld",
        },
        {
            "justAField": "sad",
            "schemaVersion": "0.2.0",
            "anotherFieldToEmbed[0].thisone": "helloworld",
        },
    ]


def test_refactor_field_props():
    flat_fields_array = [
        {"justAField": "cool", "anotherFieldToEmbed[0].thisone": "helloworld"},
        {"justAField": "sad", "anotherFieldToEmbed[0].thisone": "helloworld"},
    ]
    flat_root, flat_fields = utils.refactor_field_props(
        flat_fields_array, schema_to_rearrange
    )

    assert flat_root.to_dict() == {"anotherFieldToEmbed[0].thisone": "helloworld"}
    assert flat_fields.to_dict(orient="records") == [
        {"justAField": "cool"},
        {"justAField": "sad"},
    ]



import visions as v
import pandas.api.types as pdt
import pandas as pd
from functools import partial
# for declarative API example used as reference see:
# https://github.com/dylan-profiler/visions/blob/develop/examples/declarative_typeset.py

# for Categorical vision type reference, see ydata-profiling:
# https://github.com/ydataai/ydata-profiling/blob/develop/src/ydata_profiling/model/typeset.py
# state param needed for visions graph traversal
class contains:
    def is_category(series,state):
        return pdt.is_categorical_dtype(series)
    def is_boolean(series,state):
        return pdt.is_bool_dtype(series)

class _transformers:
    def to_category(series,state):
        return pd.Categorical(series)

class _relationships:
    def type_is_category(series,k,threshold):
        nunique = series.nunique()
        few_enough_cats = nunique<=k
        low_enough_thresh = (nunique/series.size)<threshold
        is_not_boolean = not contains.is_boolean(series,None)
        return few_enough_cats and low_enough_thresh and is_not_boolean

class inference_relations:

    def type_to_category(
        related_types=[v.Integer,v.Float,v.String],
        k=5,threshold=.2):

        relationships = []
        for vision_type in related_types:
            relation = dict(
                related_type=vision_type,
                relationship=lambda series,state: partial(_relationships.type_is_category,k=k,threshold=threshold)(series),
                transformer=_transformers.to_category)
            relationships.append(relation)
        return relationships    
# types in the CompleteSet but not in StandardSet
## just manually add additions so additional dependencies
## for unused types not necessary

# {Count,
#  Date,
#  EmailAddress,
#  File,
#  Geometry,
#  IPAddress,
#  Image,
#  Ordinal,
#  Path,
#  Time,
#  URL,
#  UUID}

#TODO: make function to allow specification of params (eg k and threshold). Currently using sensiable default (see typesets_relations.py)
#TODO: should we call this something more "frictionless-y" like "Enum"?
Categorical= v.create_type(
    "Categorical",
    identity=[v.Generic],
    contains=contains.is_category,
    inference=inference_relations.type_to_category()
)

typeset_original = (
    v.StandardSet()
    - v.Categorical
    + v.Date 
    + v.Time
)
typeset_with_categorical = (
    typeset_original
     + Categorical
 )

typeset_mapping = {
    "Integer":"integer",
    "Float":"number",
    "String":"string",
    "Time":"time",
    "TimeDelta":"duration",
    "DateTime":"datetime",
    "Boolean":"boolean"
}

def infer_frictionless_fields(
    df,
    typesets=[typeset_original,typeset_with_categorical],
    typeset_mapping=typeset_mapping):

    """ takes in a dataframe 
    and infers types by iterating through typesets.
    
    
    NOTE: This multiple typeset iteration was a solution to 
    correctly inferring types within categoricals 
    (eg integer categoricals from a float column. 
    best solution may be too look into ordering inferences but
    this does the same job.)


    """
    # TODO: infer formats with extended dtypes (eg url, email etc)
    
    for typeset in typesets:
        df,typepaths,_ = typeset.infer(df) # typepaths is list of the visions graph traversal - last item is casted type
    
    fields = []

    for col,typepath in typepaths.items():
        field = {"name":col}
        if len(typepath)==1:
            continue
        type_final = str(typepath[-1])
        if type_final=="Categorical":
            # TODO: see visions PR https://github.com/dylan-profiler/visions/issues/160 -- best way 
            # would probably be to use networkx graph. seems like this may be a good feature to add
            # to visions package
            type_second_to_final = str(typepath[-2])
            field["type"] = typeset_mapping.get(type_second_to_final,"any")
            # enums for inferred categoricals
            field["constraints"] = {"enum":list(df[col].cat.categories)}       
        else:
            field["type"] = typeset_mapping.get(str(type_final),"any")

        
        if field["type"]=="any":
            print(field["name"])
        fields.append(field)

        
    
    return fields


from ..readstat.conversion import convert_readstat,Path
import click

def convert_sas(
  input_filepath,
  data_dictionary_props,
  sas_catalog_filepath=None,
):

    ext = Path(input_filepath).suffix
    
    if ext != ".sas7bdat":
        raise Exception("Currently the SAS inputtype only support sas7bdat files")
    
    if not sas_catalog_filepath:
        sas_catalog_search = list(Path(input_filepath).parent.glob("*.sas7bcat"))
        if len(sas_catalog_search) == 1:
            sas_catalog_filepath = sas_catalog_search[0]
            click.secho(f"Using the SAS Catalog File: {str(sas_catalog_filepath)}",fg="green")
        elif len(sas_catalog_search) > 1:
            sas_catalog_filepath = sas_catalog_search[0]
            click.secho(f"Warning: Found multiple SAS Catalog files",fg="red")
            click.secho(f"Using the SAS Catalog File: {str(sas_catalog_filepath)}")
        else:
            sas_catalog_filepath = None
            click.secho("No sas catalog file found so value labels will not be applied")

    data_dictionary_package = convert_readstat(
        input_filepath,
        data_dictionary_props,
        sas_catalog_filepath=sas_catalog_filepath,
    )

    return data_dictionary_package


mapping = {'Variable / Field Name': 'name',
 'Form Name': 'form',
 'Section Header': 'section',
 'Field Type': 'type',
 'Field Label': 'label',
 'Choices, Calculations, OR Slider Labels': 'choice_calc_lbls',
 'Field Note': 'note',
 'Text Validation Type OR Show Slider Number': 'text_valid_slider_num',
 'Text Validation Min': 'text_valid_min',
 'Text Validation Max': 'text_valid_max',
 'Identifier?': 'identifier',
 'Branching Logic (Show field only if...)': 'skip_logic',
 'Required Field?': 'required',
 'Custom Alignment': 'align',
 'Question Number (surveys only)': 'question_num',
 'Matrix Group Name': 'matrix_group'}

choices_fieldname = slider_fieldname = calc_fieldname = mapping['Choices, Calculations, OR Slider Labels']
text_valid_fieldname = mapping['Text Validation Type OR Show Slider Number']



""" 
functions to read and write files in a "smart" way
"""
import pyreadstat
import pandas as pd 
from pathlib import Path
import json
import charset_normalizer
import re

from healdata_utils import schemas
# read pyreadstat files

def read_pyreadstat(file_path,**kwargs):
    ''' 
    reads in a "metadata rich file"
    (dta, sav,b7bdat). Note, xport format not supported
    as it doesnt supply value labels.

    '''
    file_path = Path(file_path)
    ext = file_path.suffix 
    if ext=='.sav':
        read = pyreadstat.read_sav
    elif ext=='.sas7bdat':
        read = pyreadstat.read_sas7bdat
    elif ext=='.dta':
        read = pyreadstat.read_dta
    elif ext=='.por':
        read = pyreadstat.read_por

    return read(file_path,**kwargs)



def detect_file_encoding(file_path):
    """ 
    detects file encoding using charset_normalizer package
    """ 
    with open(file_path,'rb') as f:
        data = f.read()
        encoding_for_input = charset_normalizer.detect(data)

    is_confident = encoding_for_input["confidence"]==1
    if not is_confident:
        print("Be careful, the detected file encoding for:")
        print(f"{file_path}")
        print(r"has less than 100% confidence")
    #chardet_normalizer.detect returns confidence,encoding (as a string), and language (eg English)
    return encoding_for_input["encoding"]


def read_delim(file_path,castdtype = "string"):
    """ 
    reads in a tabular file (ie spreadsheet) after detecting
    encoding and file extension without any type casting.

    currently supports csv and tsv

    defaults to not casting values (ie all columns are string dtypes)
    and not parsing strings into NA values (eg "" is kept as "")
    """ 
    ext = Path(file_path).suffix
    if ext==".csv":
        sep = ","
    elif ext==".tsv":
        sep = "\t"
        
    encoding = detect_file_encoding(file_path)
    file_encoding = pd.read_csv(
        file_path,sep=sep,encoding=encoding,dtype=castdtype,
        keep_default_na=False)

    return file_encoding

def read_excel(filepath,sheet_names=None,castdtype="string"):
    """ 
    reads in an excel file that outputs
    a dict of dataframes with each sheet name being
    the dict key and the dict value being the pandas 
    dataframe of the corresponding sheet. 
    
    This is akin to pandas.read_excel when all sheet_names (or sheet_name=None)
    is specified EXCEPT that the default is each value is the string representation.
    Additionally, if only one sheet exists (even if passing a list), it will output
    a datframe rather than a dict of dataframes.

    See the dtype arg in pandas read_excel docs for more info.


    file_path: str - path to xlsx file
    sheet_names: Union[str,list,None]. By default (None), all sheets are read and a dict
        of data frame is returned. If a list of sheets is provided, also returns a dict of packages. Else, returns 
    """ 

    book = pd.ExcelFile(filepath)
    
    if isinstance(sheet_names,list):
        selected_sheet_names = [sheet for sheet in book.sheet_names 
            if sheet in sheet_names]
    elif isinstance(sheet_names,(str,int)):
        selected_sheet_names = [sheet_names]
    else:
        selected_sheet_names = book.sheet_names


    if len(selected_sheet_names) == 1:
        sheet = selected_sheet_names[0]
        df = pd.read_excel(book,sheet_name=sheet,dtype=castdtype).fillna("")
        return df
    else:
        dfs = {}
        for sheet in selected_sheet_names:
            dfs[sheet] = pd.read_excel(book,sheet_name=sheet,dtype=castdtype).fillna("")
        
        return dfs


def read_zip():
    pass 


def read_archive():
    pass 

def _generate_jsontemplate(schema):

    if schema.get('type','') == 'object':
        val = {}
        if 'properties' in schema:
            for prop, prop_schema in schema['properties'].items():
                val[prop] = _generate_jsontemplate(prop_schema)
    elif schema.get('type','') == 'array':
        val = []
        if schema.get("items"):
            val.append(_generate_jsontemplate(schema['items']))
    # elif schema.get("enum"):
    #     val = f"Pick one of: {str(schema['enum'])}"
    else:
        val = None
   
    return val

    
def write_vlmd_template(outputfile,output_overwrite=False,numfields=1):

    """ 
    Writes a  json or csv template:
    If writing a json template, will iniiate null vals for all fields
    If writing a csv template (ie csv tabular template), will copy the recommendation level across num fields specified.
    

    NOTE
    -----
    - Recommendation level is in description property with expression [<recommendation level>]. The 
    "required" property takes precdence over rec level.

    - csv is a frictionless table schema and json is a jsonschema
    """  
    

    ext = Path(outputfile).suffix 

    # if file exists and overwrite option is false, raise error
    if Path(outputfile).exists() and not output_overwrite:
        raise FileExistsError(f"{outputfile} exists")


    if ext == ".json":
        schema = schemas.healjsonschema
        fields_propname = "fields"
        template = _generate_jsontemplate(schema)
        template[fields_propname] = numfields *  template[fields_propname]
        template["schemaVersion"] = schema["version"]
        Path(outputfile).write_text(json.dumps(template,indent=2))
    
    elif ext == ".csv":
        schema = schemas.healcsvschema
        vals = {}

        for propname in schema["properties"]:
            val = ""
            if propname in schema["required"]:
                val += "[Required]"
            vals[propname] = numfields * [val]
        
        for propname in schema.get("patternProperties",{}):
            newpropname = propname.replace("^","").replace("$","").replace("[\d+]","[0]")
            vals[newpropname] = numfields * [val]

        template = pd.DataFrame(vals)
        if "version" in schema:
            template["schemaVersion"] = schema["version"]

        template.to_csv(outputfile,index=False)





""" 
convert a redcap data dictionary exported
in csv format to a heal-complieant json data dictionary

""" 

import pandas as pd
from . import headers
from .mappings import typemappings 
from healdata_utils import utils
import numpy as np
from ..jsontemplate.conversion import convert_templatejson
from healdata_utils.io import read_delim
#STEPS
#1 fill section headers
#2 sort rows to allow proper row indexing
#3 make conditionals for each map function

def read(file_path):
    """ 
    Reads in a path to a redcap csv file
    and outputs and dictionary with cleaned up header (field) names
    """ 
    sourcedf = (
        read_delim(file_path)
        .fillna("")
        .rename(columns=headers.mapping)
        .map(utils.strip_html)
    )

    #downfill section (if blank -- given we read in with petl, blanks are "" but ffill takes in np.nan)
    sourcedf['section'] = sourcedf.replace({"":np.nan}).groupby('form')['section'].ffill()

    sourcedf.fillna("",inplace=True) 

    return sourcedf.to_dict(orient="records")

def gather(sourcefields):
    """ 
    maps and translates fields based on redcap field type
    to heal json
    """ 
    def __add_description(sourcefield,targetfield):

        if sourcefield.get("label"):
            fieldlabel = sourcefield["label"].strip()
        else:
            fieldlabel = ""

        if sourcefield.get("section"):
            fieldsection = sourcefield['section'].strip()+": "
        else:
            fieldsection = ""

        if targetfield.get("description"):
            fielddescription = targetfield["description"].strip()
        else:
            fielddescription = ""
        
        fielddescription = utils.strip_html((fieldsection+fieldlabel+fielddescription).strip())
        if fielddescription:
            return fielddescription
        else:
            return "No field label for this variable"

    def __add_title(sourcefield,targetfield):
        targettitle = targetfield.get("title","")

        if targettitle:
            return targettitle
        elif sourcefield.get("label"):
            return targettitle + utils.strip_html(sourcefield["label"].strip())
        else:
            return "No field label for this variable"
    
    def __add_section(sourcefield,targetfield):
        if sourcefield.get("form"):
            return sourcefield["form"]
    
    def _add_metadata(sourcefield,targetfield):
        targetfield["description"] = __add_description(sourcefield, targetfield)
        targetfield["title"] = __add_title(sourcefield, targetfield)
        targetfield["section"] = __add_section(sourcefield, targetfield)

    sourcedatafields = [field for field in sourcefields 
        if field["type"] in list(typemappings)]

    targetfields = []
    for sourcefield in sourcedatafields:
        sourcefieldtype = sourcefield["type"]
        targetfield = typemappings[sourcefieldtype](sourcefield)
        #NOTE if one sourcefield generates more than 1 target field (ie checkbox) need to iterate through
        #if list (and hence not one to one mapping with sourcefield), assumes mandatory fields
        if isinstance(targetfield,list):
            for _targetfield in targetfield:
                assert 'name' in _targetfield and 'type' in _targetfield
                _add_metadata(sourcefield,_targetfield)
                targetfields.append(_targetfield)
        else:
            _add_metadata(sourcefield,targetfield)
            targetfield_with_name = {'name':sourcefield['name']}
            targetfield_with_name.update(targetfield)
            targetfields.append(targetfield_with_name)

    return targetfields

def convert_redcapcsv(file_path,
    data_dictionary_props={}):
    """ 
    Takes in an exported Redcap Data Dictionary csv,
    and translates each field into a HEAL specified
    data dictionary based on field type (e.g., checkbox, radio, text and 
    other conditional logic based on Redcap specifications.)


    > While there are a variety of options for Redcap exports (eg directly through
    the API or via an XML), the Redcap CSV provides an easy-to-edit format comfortable by 
    technical and non-technical users.

    Parameters
    ----------
    file_path : str or path-like or an object that can be inferred as data by frictionless's Resource class.
        Data or path to data with the data being a tabular HEAL-specified data dictionary.
        This input can be any data object or path-like string excepted by a frictionless Resource object.
    data_dictionary_props : dict
        The HEAL-specified data dictionary properties.

    Returns
    -------
    dict
        A dictionary with two keys:
            - 'templatejson': the HEAL-specified JSON object.
            - 'templatecsv': the HEAL-specified tabular template.
    """ 



    sourcefields = read(file_path)
    targetfields = gather(sourcefields)

    data_dictionary = data_dictionary_props.copy()
    data_dictionary['fields'] = targetfields

    package = convert_templatejson(data_dictionary)
    return package 

    
    
    


import pandas as pd
from healdata_utils.utils import to_int_if_base10,stringify_keys
from healdata_utils.io import read_pyreadstat
from healdata_utils.types import typesets
from ..jsontemplate.conversion import convert_templatejson
from datetime import datetime
from pathlib import Path

def convert_readstat(file_path,
    data_dictionary_props=None,
    sas_catalog_filepath=None,):
    """
    Converts a "metadata-rich" (ie statistical software file) 
    into a HEAL-specified data dictionary in both csv format and json format.

    This function relies on [readstat](https://github.com/Roche/pyreadstat) which supports SPSS (sav), 
    SAS (sas7bdat), and Stata (dta). 

    > Currently, this function uses both data and metadata to generate 
    a HEAL specified data dictionary. That is, types are inferred from the 
    data (so at least test or synthetic data needed) while everything else is taken 
    from the metadata (eg missing values, variable labels, variable value labels etc)

    Parameters
    ----------
    file_path : str or path-like or any object
        Data or path to data with the data being a tabular HEAL-specified data dictionary.
        This input can be any data object or path-like string excepted by a frictionless Resource object.
    data_dictionary_props : dict
        The HEAL-specified data dictionary properties.
    
    sas_catalog_filepath : str or path-like
        Path to a sas catalog file (sas7bcat). Needed for value formats if a sas (sas7bdat) input file

    Returns
    -------
    dict
        A dictionary with two keys:
            - 'templatejson': the HEAL-specified JSON object.
            - 'templatecsv': the HEAL-specified tabular template.

    Notes
    -----
    ## Missing values (from pyreadstat docs)

    SPSS only supports 3 discrete missing in addition to ranges.
    For POC, only using discrete. TODO: use range(lo,hi+1) to do ranges; JCOIN Core Measures, for example, will need this
    
    From module documentation on missing values:

    - SPSS
        missing_ranges: a dict with keys being variable names. 
        Values are a list of dicts. 
        Each dict contains two keys, 'lo' and 'hi' being the lower boundary and higher boundary for the missing range. 
        Even if the value in both lo and hi are the same, the two elements will always be present. 
        This appears for SPSS (sav) files when using the option user_missing=True: user defined missing values appear not as nan but as their true value and this dictionary stores the information about which values are to be considered missing.
    
    - Stata/SAS
        missing_user_values: a dict with keys being variable names. 
        Values are a list of character values (A to Z and _ for SAS, a to z for SATA) 
        representing user defined missing values in SAS and STATA. 
        This appears when using user_missing=True in read_sas7bdat or read_dta 
        if user defined missing values are present.

    """
    metaparams = dict(file_path=file_path,user_missing=True)
    if sas_catalog_filepath:
        metaparams["catalog_file"] = sas_catalog_filepath
    _,meta = read_pyreadstat(**metaparams) # get user missing values (for stata/sas will make string so need sep call to infer types)
    df,_ = read_pyreadstat(file_path) # dont fill user defined missing vals (to get correct types)
    
    if not data_dictionary_props:
        data_dictionary_props = {}

    # NOTE: not inferring categoricals as this is inferred from value labels
    fields = typesets.infer_frictionless_fields(df,typesets=[typesets.typeset_original])

    for field in fields:
        field.pop('extDtype',None)
        fieldname = field['name']

        variable_label = meta.column_names_to_labels.get(fieldname)
        if variable_label:
            field['description'] = variable_label

        missing_values = meta.missing_user_values.get(fieldname,[])
        missing_ranges = meta.missing_ranges.get(fieldname,[])
        #see NOTE in docstring (on missing values): 
        # below maps SPSS missing values
        for items in missing_ranges:
            values = list(set(items.values()))
            if len(values)==1:
                if isinstance(values[0],datetime):
                    missing_values.append(str(values[0]))
                else:
                    missing_values.append(values[0])
            # missing_ranges is in readstat for spss
            elif list(items.keys()) == ["lo","hi"]:
                values = list(range(int(items["lo"]),int(items["hi"])+1))
                missing_values.extend(values)
            else:
                raise Exception("Currently, only discrete values are supported")

        # NOTE: stringify to conform to frictionless pattern
        value_labels = meta.variable_value_labels.get(fieldname)

        if value_labels:
            # NOTE: collect enums before stringfying - frictionless spec evaluates on logical representation
            enums = [
                val for val in list(value_labels.keys()) 
                if not val in missing_values
            ]
            stringify_keys(value_labels)
            field['enumLabels'] = value_labels

            if enums:
                constraints_enums = {'constraints':{'enum':enums}}
                field.update(constraints_enums)
        
        if missing_values:
            # NOTE: stringify to conform to frictionless standards (missing values evaluated on physical representation)
            missing_values_str = [str(val) for val in missing_values]
            field['missingValues'] = missing_values_str

    data_dictionary = data_dictionary_props.copy()
    data_dictionary['fields'] = fields 

    package = convert_templatejson(data_dictionary)
    return package



''' 
takes the filled out variable level metadata template
and converts to json, validates, and saves json template to file
based on an explicit set of mapping functions

goes the other way (from json to csv as well)

see convert_templatecsv_to_json and convert_json_to_templatecsv
''' 
import petl as etl
import pandas as pd
from pathlib import Path
import collections
# from frictionless import Resource,Package
from healdata_utils.io import read_delim
from healdata_utils import schemas,utils,mappings
from healdata_utils.mappings import versions
from os import PathLike
import re
def convert_datadictcsv(
    csvtemplate: str,
    data_dictionary_props: dict,
    renamemap:dict=None,
    recodemap:dict=None,
    droplist:dict=None,
    item_sep:str="|", 
    keyval_sep:str="=",
    **kwargs
) -> dict:
    """
    [shortdesc]
    Converts a CSV conforming to HEAL specifications (but see 2 additional notes below) 
    into a HEAL-specified data dictionary in both csv format and json format.

    Converts an in-memory data dictionary or a path to a data dictionary file into a HEAL-specified tabular template by:
        1. Adding missing fields, and
        2. Converting fields from a specified mapping.
            NOTE: currently this mapping is only float/num to number or text/char to string (case insensitive)
                In future versions, there will be a specified module for csv input mappings.
    [shortdesc]
    
    Parameters
    ----------
    csvtemplate : str or path-like or an object that can be inferred as data by frictionless's Resource class.
        Data or path to data with the data being a tabular HEAL-specified data dictionary.
        This input can be any data object or path-like string excepted by a frictionless Resource object.
    data_dictionary_props : dict
        The HEAL-specified data dictionary properties.
    renamemap: A mapping of source (current) column headers to target (desired -- conforming to CVS HEAL spec)
        column headers 
    recodemap: A mapping of values for each column in HEAL spec -- {..."colname":{"oldvalue":"newvalue"...}...}
    droplist: a list of variables to drop from headers before processing
    item_sep:str (default:"|") Used to split stringified items (in objects and arrays)
    keyval_sep:str (default:"=") Used to split stringified each key-value pair

    Returns
    -------
    dict
        A dictionary with two keys:
            - 'templatejson': the HEAL-specified JSON object.
            - 'templatecsv': the HEAL-specified tabular template.

    """
    # get transforms
    # cast numbers explicitly based on schema
    # this is needed in case there is only one record in a string column that is a number (ie don't want to convert)
    def infer_delim(series:pd.Series,char_list:list,firstmatch:bool):
        """ infer the delimiter by the highest frequency by row higest occuring character
        can either take only the first match or all matches for each row""" 
        combined_pattern = "|".join(map(re.escape,char_list))

        try:
            # get most frequent delimiter per row
            most_freq_chars = (
                series
                .rename_axis("row")
                .str.extractall(f"({combined_pattern})")
                .squeeze("rows") # only 1 row, dont want to squeeze to scalar so spec rows
                .pipe(lambda s: s.unstack("match")[0] if firstmatch else s) # if first match specified
                .groupby("row")
                .value_counts()
                .unstack()
                .idxmax(axis=1)
            )
        except KeyError:
            return None
        
        try:
            # infer delim by character with most per row max frequencies
            inferred_delim = (
                most_freq_chars
                .value_counts()
                .sort_values()
                .index
                [-1]
            )
        except IndexError:
            return None

        return inferred_delim




    if isinstance(csvtemplate,(str,PathLike)):
        template_tbl = read_delim(str(Path(csvtemplate)))
    else:
        template_tbl = pd.DataFrame(csvtemplate)

    if not renamemap:
        renamemap = {}

    if not recodemap:
        recodemap = {}

    if not droplist:
        droplist = []

    slugify = lambda s: s.strip().lower().replace("_","-").replace(" ","-") 
    # flattened properties
    field_properties = utils.flatten_properties(
        schemas.healjsonschema
        ["properties"]
        ["fields"]
        ["items"]
        ["properties"]
    )
    
    # init to-be formatted tables
    tbl_csv = template_tbl.copy()
    # transform each column with slugified mappings, harmonizing delims (if array, object) 
    for colname in tbl_csv.columns.tolist():
        slugified_col = slugify(colname)
        newcolname = colname

        # rename based on slugified names or original col names
        newcolname = renamemap.get(slugified_col) or renamemap.get(colname)
        if newcolname:
            tbl_csv.rename(columns={colname:newcolname},inplace=True)
        else:
            newcolname = colname
        
        # recode slugified old names to accepted names
        if newcolname in recodemap.keys():
            tbl_csv[newcolname] = (
                tbl_csv[newcolname]
                .apply(slugify)
                .replace(recodemap[newcolname])
            )

        if newcolname in droplist:
            del tbl_csv[newcolname]

        # NOTE: the below methodology uses the schema to instruct how to convert to json.
        fieldpropname = utils.find_propname(newcolname,field_properties)
        fieldprop = field_properties.get(fieldpropname)
        # infer delimiters of stringified lists (note: stringified lists are identified from schema)

        
        if fieldpropname:
            
            if fieldprop["type"] == "integer":
                tbl_csv[newcolname] = tbl_csv[newcolname].apply(lambda s: int(float(s)) if s else s)
            elif fieldprop["type"] == "number":
                tbl_csv[newcolname] = tbl_csv[newcolname].astype(float)
            elif fieldprop["type"] == "object":

                possible_keyval = ["=",":"]
                possible_list = [";","|"]
                keyval_sep = infer_delim(tbl_csv[newcolname],possible_keyval,firstmatch=True) or "="
                item_sep = infer_delim(tbl_csv[newcolname],possible_list,firstmatch=False) or "|"

                tbl_csv[newcolname] = (
                    tbl_csv[newcolname]
                    .str.replace(keyval_sep,"=")
                    .str.replace(item_sep,"|")
                )

            elif fieldprop["type"] == "array":
                possible_list = [";","|"]
                item_sep = infer_delim(tbl_csv[newcolname],possible_list,firstmatch=False) or "|"
                tbl_csv[newcolname] = tbl_csv[newcolname].replace(item_sep,"|")

    # parse string objects and array to create the dict (json) instance
    tbl_json = tbl_csv.copy()   
    for colname in tbl_json.columns.tolist():
        # NOTE: the below methodology uses the schema to instruct how to convert to json.
        fieldpropname = utils.find_propname(colname,field_properties)
        fieldprop = field_properties.get(fieldpropname)
        if fieldprop:
            if fieldprop["type"] == "object":
                tbl_json[colname] = tbl_csv[colname].apply(
                    utils.parse_dictionary_str,
                    item_sep="|",keyval_sep="=")
            
            elif fieldprop["type"] == "array":
                tbl_json[colname] = tbl_csv[colname].apply(
                    utils.parse_list_str,item_sep="|")
        # columns not included in schema (custom or other)
        else:
            if colname.split(".")[0] == "custom":
                
                if not "custom" in tbl_json:
                    tbl_json["custom"] = [{}] * len(tbl_json)

                for i in range(len(tbl_csv)):
                    value = tbl_csv[colname].iloc[i]
                    if value:
                        custom = {}
                        parts = colname.split(".")[1:]
                        
                        for key in reversed(parts):
                            result = {key: value}

                        tbl_json["custom"].iloc[i].update(result)
            else:
                tbl_json[colname] = tbl_csv[colname]

    # drop all custom columns (as I have nested already)
    tbl_json.drop(columns=tbl_json.filter(regex="^custom\\.").columns, inplace=True)

    # refactor (i.e., cascade, move up to root) properties if present in all records
    refactored_props,tbl_json = utils.refactor_field_props(
        tbl_json,schema=schemas.healjsonschema)

    # data dictionary root level properties
    data_dictionary_props_csv = dict(data_dictionary_props)
    data_dictionary_props_json = {**data_dictionary_props,**utils.unflatten_from_jsonpath(refactored_props.to_dict())}

    # create the data dictionary objects
    fields_json = [utils.unflatten_from_jsonpath(record) 
        for record in tbl_json.to_dict(orient="records")]
    template_json = dict(**data_dictionary_props_json,fields=fields_json)

    fields_csv = tbl_csv.to_dict(orient="records")
    template_csv = dict(**data_dictionary_props_csv,fields=fields_csv)

    return {"templatejson":template_json,"templatecsv":template_csv}


from pathlib import Path
import shutil
import json

from healdata_utils.io import write_vlmd_template,read_excel

from conftest import _compare_vlmd_tmp_to_output,fields_propname

def test_write_vlmd_template():



    tmpdir = Path("tmp")

    if tmpdir.exists():
        shutil.rmtree(tmpdir)

    tmpdir.mkdir(exist_ok=True)

    write_vlmd_template(tmpdir.joinpath("heal.json"),numfields=2)
    write_vlmd_template(tmpdir.joinpath("heal.csv"),numfields=2)


    csvoutput = Path("data/templates/twofields.csv").read_text().split("\n")
    jsonoutput = json.loads(Path("data/templates/twofields.json").read_text())

    _compare_vlmd_tmp_to_output(tmpdir/"heal",csvoutput,jsonoutput,fields_propname)

    shutil.rmtree(tmpdir)


# def test_read_excel():
    
#     excelfile = read_excel("data/misc/multitab-data-dictionary.xlsx")



""" 
Functions for each redcap field type
that uses redcap field info to determine
various pieces of field metadata.

Input assumes a dictionary with key=redcap fieldname;value=redcap value
""" 
import re
from healdata_utils import utils
from .headers import choices_fieldname,slider_fieldname,calc_fieldname,text_valid_fieldname 
# NOTES	- large text box for lots of text
# DROPDOWN	- dropdown menu with options
# RADIO	- radio buttons with options
# CHECKBOX	- checkboxes to allow selection of more than one option
# FILE	- upload a document
# CALC	- perform real-time calculations
# SQL	- select query statement to populate dropdown choices
# DESCRIPTIVE	- text displayed with no data entry and optional image/file attachment
# SLIDER	- visual analogue scale; coded as 0-100
# YESNO	- radio buttons with yes and no options; coded as 1, Yes | 0, No
# TRUEFALSE	- radio buttons with true and false options; coded as 1, True | 0, False

def _parse_field_properties_from_encodings(
    encodings_string
    ):  
    """ 
    Many of the field types have the same logic
    for conversion to data types and just differ
    in presentation (dropbox,radio box) so making
    fxn to support this

    Currently supports strings,ints,and nums for types

    NOTE: this function may be best housed in the general 
    utils but keeping local for now and specific to 
    redcap delimiters
    """ 
    # parse encodings
    fieldencodings = utils.parse_dictionary_str(
        encodings_string, item_sep="|", keyval_sep=",")
    # get enums
    fieldenums = list(fieldencodings.keys())
    #interpret type from enums
    for val in fieldenums:
        val = val.strip()
        if val.isnumeric():
            try:
                int(val)
                fieldtype = "integer"
            except ValueError:
                fieldtype = "number"
        else:
            fieldtype = "string"
    
    return {
        "type":fieldtype,
        "enumLabels":{key.strip():val.strip() for key,val in fieldencodings.items()},
        "constraints":{
            "enum":[val.strip() for val in fieldenums]
        }
    }

def maptext(field):
    """ 
    TEXT - single-line text box (for text and numbers)

    looks at text validation field
    """
    if field.get(text_valid_fieldname):
        text_validation = field[text_valid_fieldname].lower()
    else:
        text_validation = ""
    fieldformat = None 
    fieldpattern = None
    if "datetime" in text_validation:
        fieldtype = "datetime"
        fieldformat = "any"
    elif "date" in text_validation:
        fieldtype = "date"
        fieldformat = "any" 
    elif text_validation=="email":
        fieldtype = "string"
        fieldformat = "email"
    elif text_validation=="integer":
        fieldtype = "integer"
    elif text_validation=="alpha_only":
        fieldtype = "string"
        fieldpattern = "^[a-zA-Z]+$"
    elif "number" in text_validation:
        fieldtype = "number"
        if "comma_decimal" in text_validation:
            fielddecimal_char = ","
    elif text_validation=="phone":
        fieldtype = "string"
        fieldpattern = "^[0-9]{3}-[0-9]{3}-[0-9]{4}$" 
    elif text_validation=="postalcode_australia":
        fieldtype = "string"
        fieldpattern = "^[0-9]{4}$"
    elif text_validation=="postalcode_canada":
        fieldtype = "string"
        fieldpattern = "^[A-Z][0-9][A-Z] [0-9][A-Z][0-9]$"
    elif text_validation=="ssn":
        fieldtype = "string"
        fieldpattern = "^[0-9]{3}-[0-9]{2}-[0-9]{4}$"
    elif "time" in text_validation:
        fieldtype = "time"
        fieldformat = "any"
    elif text_validation=="vmrn":
        fieldtype = "string"
        fieldpattern = "^[0-9]{10}$"
    elif text_validation=="zipcode":
        fieldtype = "string"
        fieldpattern = "^[0-9]{5}$"
    else:
        fieldtype = "string"
    
    props = dict(zip(
        ["type","format","constraints"],
        [fieldtype,fieldformat,{"pattern":fieldpattern}]
    ))

    # delete props without values (ie None)
    for propname in ["type","format"]:
        if props[propname]==None:
            del props[propname]
    if props["constraints"]["pattern"]==None:
        del props["constraints"]

    return props
def mapnotes(field):
    """ NOTES	
    large text box for lots of text
    """
    return {"type":"string"}

def mapdropdown(field):
    """ 
    DROPDOWN	
    dropdown menu with options

    Determined by "options" (ie Choices, Calculations, OR Slider Labels)
    """
    encodings_string = field[choices_fieldname]
    return _parse_field_properties_from_encodings(encodings_string)

def mapradio(field):
    """ 
    RADIO	- radio buttons with options

    Determined by "options" (ie Choices, Calculations, OR Slider Labels)

    """
    # parse enum/encodings 
    encodings_string = field[choices_fieldname]
    return _parse_field_properties_from_encodings(encodings_string)


def mapcheckbox(field):
    """ 
    CHECKBOX	- checkboxes to allow selection of more than one option


    ## Are data from checkbox (choose all that apply) field types handled differently from other field types when imported or exported?
    Yes. When your data are exported, each option from a checkbox field becomes a separate variable coded 1 or 0 to reflect whether it is checked or unchecked. By default, each option is pre-coded 0, so even if you have not yet collected any data, you will see 0's for each checkbox option. The variable names will be the name of the field followed by the option number. So, for example, if you have a field coded as follows:

    Race

    1, Caucasian

    2, African American

    3, Asian

    4, Other

    In your exported dataset, you will have four variables representing the field Race that will be set as 0 by default, coded 1 if the option was checked for a record. The variable names will consist of the field name. three underscores, and the choice value:

    race___1
    race___2
    race___3
    race___4

    Notes:

    when you import data into a checkbox field, you must code it based on the same model
    negative values can be used as the raw coded values for checkbox fields. Due to certain limitations, negative values will not work when importing values using the Data Import Tool, API and cause problems when exporting data into a statistical analysis package. The workaround is that negative signs are replaced by an underscore in the export/import-specific version of the variable name (e.g., for a checkbox named "race", its choices "2" and "-2" would export as the fields
    race___2

    race____2

    A checkbox field can be thought of as a series of yes/no questions in one field. Therefore, a yes (check) is coded as 1 and a no (uncheck) is coded a 0. An unchecked response on a checkbox field is still regarded as an answer and is not considered missing.
    """
    checkboxname = field['name']
    choices = utils.parse_dictionary_str(
        field[choices_fieldname], item_sep="|", keyval_sep=",")
    fieldtype = "boolean"
    fieldenums = ["0","1"]
    fieldencodings = {"0":"Unchecked","1":"Checked"}

    fieldsnew = [
        {
            "description":f"[choice={choice}]",
            "title": checkboxname.title()+": "+choice,
            "name":checkboxname+"___"+re.sub("^\-","_",val).strip(), #NOTE: REDCAP changes negative sign to underscore
            "type":fieldtype,
            "constraints":{"enum":fieldenums},
            "enumLabels":fieldencodings
        }
        for val,choice in choices.items()
    ]
    return fieldsnew

def mapfile(field):
    return {"type":"string"}


def mapcalc(field):
    return {
        "description":f"[calculation: {field[calc_fieldname]}]",
        "type": "number"
    }


def mapsql(field):
    return None


def mapyesno(field):
    return {
        "type":"boolean",
        "constraints":{"enum":["0","1"]},
        "enumLabels":{"0":"No","1":"Yes"}
    }


def maptruefalse(field):
    return {
        "type":"boolean",
        "constraints":{"enum":["0","1"]},
        "enumLabels":{"0":"False","1":"True"}
    }


def mapslider(field):
    vallist = ["0","50","100"]
    lbllist = utils.parse_list_str(field[slider_fieldname],"|") 
    fieldencodings = {vallist[i]:lbl for i,lbl in enumerate(lbllist)}
    return {
        "type":"integer",
        "constraints":{
            "minimum":0,
            "maximum":100
        },
        "enumLabels":fieldencodings
    }
def mapdescriptive(field):
    return None


#not mapping descriptives or sql (TODO: mapping sql?)
typemappings = {
    "text":maptext,
    "notes":mapnotes,
    "dropdown":mapdropdown,
    "radio":mapradio,
    "checkbox":mapcheckbox,
    "slider":mapslider,
    "yesno":mapyesno,
    "truefalse":maptruefalse,
    "calc":mapcalc,
    "file":mapfile
}





""" 
validation of json records against a jsonschema specification 
and conversion of one schema spec to jsonschema specification 
eg frictionless --> jsonschema

""" 

import jsonschema

def validate_against_jsonschema(json_object,schema):
    
    Validator = jsonschema.validators.validator_for(schema)
    validator = Validator(schema)
    report = []
    is_valid = True
    for error in validator.iter_errors(json_object):
        is_valid = False
        error_report = {
            "json_path":error.json_path,
            "message":error.message,
            "absolute_path":list(error.path),
            "relative_path":list(error.relative_path),
            "validator":error.validator,
            "validator_value":error.validator_value
        }
        report.append(error_report)

    return {"valid":is_valid,"errors":report}


import shutil
from pathlib import Path
from healdata_utils.conversion import convert_to_vlmd,input_short_descriptions,choice_fxn
import json

from conftest import compare_vlmd_tmp_to_output
from conftest import valid_input_params,valid_output_json,valid_output_csv,fields_propname

def test_convert_to_vlmd_with_registered_formats():

    outputdir="tmp"
    for inputtype,_valid_input_params in valid_input_params.items():
   
        # make an empty temporary output directory
        try:
            Path(outputdir).mkdir()
        except FileExistsError:
            shutil.rmtree(outputdir)
            Path(outputdir).mkdir()

        data_dictionaries = convert_to_vlmd(**_valid_input_params)
        compare_vlmd_tmp_to_output(_valid_input_params)
    
        # clean up
        shutil.rmtree(outputdir)

def test_short_descriptions():
    
    assert input_short_descriptions.keys() == choice_fxn.keys()


from pathlib import Path
import json
# from frictionless import Resource,Package
import collections
from healdata_utils import utils,schemas
from os import PathLike
import pandas as pd

def convert_templatejson(
    jsontemplate,
    data_dictionary_props:dict=None,
    fields_name:str='fields',
    sep_iter = '|',
    sep_dict = '=',
    **kwargs
    ):
    """
    Converts a JSON file or dictionary conforming to HEAL specifications
    into a HEAL-specified data dictionary in both csv format and json format.

    Converts in-memory data or a path to a data dictionary file.

    If data_dictionary_props is specified, any properties passed in will be
    overwritten.
    
    Parameters
    ----------
    jsontemplate : str or path-like or an object that can be inferred as data by frictionless's Resource class.
        Data or path to data with the data being a tabular HEAL-specified data dictionary.
        This input can be any data object or path-like string excepted by a frictionless Resource object.
    data_dictionary_props : dict
        The HEAL-specified data dictionary properties.

    Returns
    -------
    dict
        A dictionary with two keys:
            - 'templatejson': the HEAL-specified JSON object.
            - 'templatecsv': the HEAL-specified tabular template.


    TODO
    ---------

    Allow an array of fields to be passed in

    """
    if isinstance(jsontemplate,(str,PathLike)):
        jsontemplate_dict = json.loads(Path(jsontemplate).read_text())
    elif isinstance(jsontemplate, collections.abc.MutableMapping):
        jsontemplate_dict = jsontemplate
    else:
        raise Exception("jsontemplate needs to be either dictionary-like or a path to a json")

    if data_dictionary_props:
        for propname,prop in data_dictionary_props.items():

            # determine if you should write or overwrite the
            ## root level data dictionary props
            if not jsontemplate_dict.get(propname):
                write_prop = True
            elif prop and prop!=jsontemplate_dict.get(propname):
                write_prop = True
            else:
                write_prop = False

            if write_prop:
                jsontemplate_dict[propname] = prop

    fields_json = jsontemplate_dict.pop(fields_name)
    data_dictionary_props = jsontemplate_dict

    fields_schema = schemas.healjsonschema["properties"]["fields"]["items"]
    flattened_fields = pd.DataFrame([utils.flatten_to_jsonpath(f,fields_schema) 
            for f in fields_json])
    flattened_data_dictionary_props = pd.Series(utils.flatten_to_jsonpath(data_dictionary_props,schemas.healjsonschema))

    flattened_and_embedded = utils.embed_data_dictionary_props(flattened_fields,flattened_data_dictionary_props,schemas.healjsonschema)
    tbl_csv = (
        flattened_and_embedded
        .fillna("")
        .map(lambda v: utils.join_dictitems(v) if isinstance(v,collections.abc.MutableMapping) else v)
        .map(lambda v: utils.join_iter(v) if isinstance(v,collections.abc.MutableSequence) else v)
    )
    fields_csv = tbl_csv.to_dict(orient="records")

    template_json = {**data_dictionary_props,"fields":fields_json}
    template_csv = {**data_dictionary_props,"fields":fields_csv}

    return {"templatejson":template_json,"templatecsv":template_csv}


from click.testing import CliRunner
from healdata_utils.cli import vlmd
from conftest import _compare_vlmd_tmp_to_output,compare_vlmd_tmp_to_output
from conftest import valid_input_params,valid_output_json,valid_output_csv,fields_propname
import shutil
import os
import json
from pathlib import Path

def test_vlmd_extract_all_params():

    for inputtype,_input_params in valid_input_params.items():

        # collect CLI arguments
        cli_params = ['extract']
        
        for paramname,param in _input_params.items():
            
            # add CLI options
            if paramname=="output_filepath":
                cli_params.append("--outputfile")
                cli_params.append(str(param))
            elif paramname=="input_filepath":
                cli_args = str(param)  # click argument
            elif paramname == "data_dictionary_props":
                for _paramname,_param in param.items():
                    cli_params.append("--prop")
                    cli_params.append(_paramname)
                    cli_params.append(_param)
            elif paramname == "sas_catalog_filepath":
                # CLI currently infers sas catalog file
                pass

            elif paramname in ["inputtype"]:
                cli_params.append(f"--{paramname.replace('_','-')}")
                cli_params.append(str(param))

        # make an empty temporary output directory
        _outdir = _input_params["output_filepath"].parent
        try:
            Path(_outdir).mkdir()
        except FileExistsError:
            shutil.rmtree(_outdir)
            Path(_outdir).mkdir()

        # add click arguments at end
        cli_params.append(cli_args)

        #run CLI
        runner = CliRunner()
        result = runner.invoke(vlmd, cli_params)

        assert result.exit_code == 0

        compare_vlmd_tmp_to_output(_input_params)
        # clean up
        shutil.rmtree(_outdir)


def test_vlmd_extract_minimal():

    for testname,testparams in valid_input_params.items():

        filepath = str(Path(testparams["input_filepath"]).resolve())
        inputtype = testparams["inputtype"]

        try:
            Path("tmp").mkdir()
        except FileExistsError:
            shutil.rmtree("tmp")
            Path("tmp").mkdir()

        os.chdir("tmp")

        # collect CLI arguments
        
        cli_params = ['extract',"--inputtype",inputtype,filepath]

        #run CLI
        runner = CliRunner()
        result = runner.invoke(vlmd, cli_params)

        assert result.exit_code == 0,result.output


        # clean up
        os.chdir("..")
        shutil.rmtree("tmp")


def test_vlmd_validate():

    paths = Path("data/valid/output").rglob("*")
    for path in paths:
        if path.is_file():
            runner = CliRunner()
            result = runner.invoke(vlmd, ['validate',str(path)])

            assert result.exit_code == 0,result.output


def test_vlmd_template():

    tmpdir = Path("tmp")

    if tmpdir.exists():
        shutil.rmtree(tmpdir)

    tmpdir.mkdir()

    runner = CliRunner()
    resultjson = runner.invoke(vlmd, ['template',"tmp/templatejson.json","--numfields","2"])
    resultcsv = runner.invoke(vlmd, ['template',"tmp/templatecsv.csv","--numfields","2"])

    assert resultjson.exit_code == 0,resultjson.output
    assert resultcsv.exit_code == 0,resultcsv.output

    # need to rename output files as the compare fxn uses same stem for csv and json
    # however, overwrite check looks for just the stem so need diff names in above fxn
    os.rename("tmp/templatejson.json","tmp/template.json")
    os.rename("tmp/templatecsv.csv","tmp/template.csv")

    csvoutput = Path("data/templates/twofields.csv").read_text().split("\n")
    jsonoutput = json.loads(Path("data/templates/twofields.json").read_text())

    _compare_vlmd_tmp_to_output("tmp/template",csvoutput,jsonoutput,fields_propname)

    shutil.rmtree(tmpdir)


if __name__=="__main__":
    vlmd.main()



healjsonschema = {
    "version": "0.3.2",
    "$schema": "http://json-schema.org/draft-07/schema#",
    "$id": "vlmd",
    "title": "Variable Level Metadata (Data Dictionaries)",
    "description": "This schema defines the variable level metadata for one data "
    "dictionary for a given study.Note a given study can have "
    "multiple data dictionaries.",
    "type": "object",
    "required": ["title", "fields"],
    "additionalProperties": False,
    "properties": {
        "title": {"type": "string"},
        "description": {"type": "string"},
        "schemaVersion": {
            "type": "string",
            "description": "The version of the schema "
            "used in agreed upon "
            "convention of "
            "major.minor.path (e.g., "
            "1.0.2) \n"
            "\n"
            "NOTE: This is NOT for "
            "versioning of each "
            "indiviual data dictionary "
            "instance. \n"
            "Rather, it is the\n"
            "version of THIS schema "
            "document. See `version` "
            "property (below) if "
            "specifying the individual "
            "data dictionary instance\n"
            "version.\n"
            "\n"
            "If generating a vlmd "
            "document as a csv file, "
            "include this version in \n"
            "every row/record to "
            "indicate this is a schema "
            "level property \n"
            "(not applicable for the "
            "json version as this "
            "property is already at "
            "the schema/root level)\n",
            "pattern": "\\d+\\.\\d+\\.\\d+",
            "examples": ["1.0.0", "0.2.0"],
        },
        "version": {
            "type": "string",
            "description": "The specified individual data "
            "dictionary instance version.",
        },
        "standardsMappings": {
            "type": "array",
            "description": "A set of standardized "
            "instruments linked to "
            "all variables within "
            "the `fields` property "
            "(but see note).\n"
            "\n"
            '!!! note "NOTE"\n'
            "\n"
            "  If "
            "`standardsMappings` "
            "is present at both "
            "the root (this "
            "property) and within "
            "`fields`, \n"
            "  then the `fields` "
            "`standardsMappings` "
            "property takes "
            "precedence.\n"
            "\n"
            "  Note, only "
            "instrument can be "
            "mapped to this "
            "property as opposed "
            "to the `fields` "
            "`standardsMappings`\n"
            "  This property has "
            "the same "
            "specification as the "
            "`fields` "
            "`standardsMappings` "
            "to make the cascading "
            "logic\n"
            "  easier to "
            "understand in the "
            "same way other "
            "standards implement "
            "cascading \n"
            "  (e.g., "
            "`missingValues` in "
            "the [frictionless "
            "specification](https://specs.frictionlessdata.io/patterns/#missing-values-per-field))\n",
            "items": {
                "properties": {
                    "type": "object",
                    "instrument": {
                        "type": "object",
                        "title": "Standard " "mapping " "- " "instrument",
                        "description": "A "
                        "standardized "
                        "set "
                        "of "
                        "items "
                        "which "
                        "encompass \n"
                        "a "
                        "variable "
                        "in "
                        "this "
                        "variable "
                        "level "
                        "metadata "
                        "document "
                        "(if "
                        "at "
                        "the "
                        "root "
                        "level "
                        "or "
                        "the "
                        "document "
                        "level) \n"
                        "or "
                        "the "
                        "individual "
                        "variable "
                        "(if "
                        "at "
                        "the "
                        "field "
                        "level). \n"
                        "\n"
                        "\n"
                        "!!! "
                        "note "
                        '"NOTE"\n'
                        "\n"
                        "  "
                        "If "
                        "information "
                        "is "
                        "present "
                        "at "
                        "both "
                        "the "
                        "root "
                        "and "
                        "the "
                        "field "
                        "level, \n"
                        "  "
                        "then "
                        "the "
                        "information "
                        "at "
                        "the "
                        "field "
                        "level "
                        "would "
                        "take "
                        "precedence "
                        "(i.e., "
                        "it "
                        "would "
                        "cascade).\n",
                        "properties": {
                            "url": {
                                "title": "Url",
                                "description": "A "
                                "url "
                                "(e.g., "
                                "link, "
                                "address) "
                                "to "
                                "a "
                                "file "
                                "or "
                                "other "
                                "resource "
                                "containing "
                                "the "
                                "instrument, "
                                "or\n"
                                "a "
                                "set "
                                "of "
                                "items "
                                "which "
                                "encompass "
                                "a "
                                "variable "
                                "in "
                                "this "
                                "variable "
                                "level "
                                "metadata "
                                "document "
                                "(if "
                                "at "
                                "the "
                                "root "
                                "level "
                                "or "
                                "the "
                                "document "
                                "level) \n"
                                "or "
                                "the "
                                "individual "
                                "variable "
                                "(if "
                                "at "
                                "the "
                                "field "
                                "level). \n",
                                "type": "string",
                                "format": "uri",
                                "examples": [
                                    "https://www.heal.nih.gov/files/CDEs/2023-05/adult-demographics-cdes.xlsx"
                                ],
                            },
                            "source": {
                                "type": "string",
                                "title": "Source",
                                "description": "An "
                                "abbreviated "
                                "name/acronym "
                                "from "
                                "a "
                                "controlled "
                                "vocabulary "
                                "referencing "
                                "the "
                                "resource "
                                "(e.g., "
                                "program "
                                "or "
                                "repository)\n"
                                "containing "
                                "the "
                                "instrument, "
                                "or "
                                "a "
                                "set "
                                "of "
                                "items "
                                "which "
                                "encompass "
                                "a "
                                "variable "
                                "in "
                                "this "
                                "variable "
                                "level "
                                "metadata "
                                "document "
                                "(if "
                                "at "
                                "the "
                                "root "
                                "level "
                                "or "
                                "the "
                                "document "
                                "level) \n"
                                "or "
                                "the "
                                "individual "
                                "variable "
                                "(if "
                                "at "
                                "the "
                                "field "
                                "level). \n",
                                "enum": ["heal-cde"],
                            },
                            "title": {
                                "type": "string",
                                "title": "Title",
                                "examples": [
                                    "Adult " "demographics",
                                    "adult-demographics",
                                ],
                            },
                            "id": {
                                "type": "string",
                                "title": "Identifier",
                                "description": "A "
                                "code "
                                "or "
                                "other "
                                "string "
                                "that "
                                "identifies "
                                "the "
                                "instrument "
                                "within "
                                "the "
                                "source.\n"
                                "This "
                                "should "
                                "always "
                                "be "
                                "from "
                                "the "
                                "source's "
                                "formal, "
                                "standardized "
                                "identification "
                                "system \n",
                                "examples": ["5141"],
                            },
                        },
                    },
                }
            },
        },
        "custom": {
            "type": "object",
            "description": "Additional properties not "
            "included as a core property. \n",
        },
        "fields": {
            "type": "array",
            "items": {
                "title": "HEAL Variable Level " "Metadata Fields",
                "description": "\n"
                "\n"
                '!!! note "Highly '
                'encouraged"\n'
                "\n"
                "  - Only `name` and "
                "`description` "
                "properties are "
                "required. \n"
                "  - For categorical "
                "variables, "
                "`constraints.enum` "
                "and `enumLabels` "
                "(where applicable) "
                "properties are highly "
                "encouraged. \n"
                "  - For studies using "
                "HEAL or other common "
                "data elements (CDEs), "
                "`standardsMappings` "
                "information is highly "
                "encouraged.\n"
                "  - `type` and "
                "`format` properties "
                "may be particularly "
                "useful for some "
                "variable types (e.g. "
                "date-like "
                "variables)\n",
                "type": "object",
                "required": ["name", "description"],
                "additionalProperties": False,
                "properties": {
                    "schemaVersion": {
                        "type": "string",
                        "description": "The "
                        "version "
                        "of "
                        "the "
                        "schema "
                        "used "
                        "in "
                        "agreed "
                        "upon "
                        "convention "
                        "of "
                        "major.minor.path "
                        "(e.g., "
                        "1.0.2) \n"
                        "\n"
                        "NOTE: "
                        "This "
                        "is "
                        "NOT "
                        "for "
                        "versioning "
                        "of "
                        "each "
                        "indiviual "
                        "data "
                        "dictionary "
                        "instance. \n"
                        "Rather, "
                        "it "
                        "is "
                        "the\n"
                        "version "
                        "of "
                        "THIS "
                        "schema "
                        "document. "
                        "See "
                        "`version` "
                        "property "
                        "(below) "
                        "if "
                        "specifying "
                        "the "
                        "individual "
                        "data "
                        "dictionary "
                        "instance\n"
                        "version.\n"
                        "\n"
                        "If "
                        "generating "
                        "a "
                        "vlmd "
                        "document "
                        "as "
                        "a "
                        "csv "
                        "file, "
                        "include "
                        "this "
                        "version "
                        "in \n"
                        "every "
                        "row/record "
                        "to "
                        "indicate "
                        "this "
                        "is "
                        "a "
                        "schema "
                        "level "
                        "property \n"
                        "(not "
                        "applicable "
                        "for "
                        "the "
                        "json "
                        "version "
                        "as "
                        "this "
                        "property "
                        "is "
                        "already "
                        "at "
                        "the "
                        "schema/root "
                        "level)\n",
                        "pattern": "\\d+\\.\\d+\\.\\d+",
                        "examples": ["1.0.0", "0.2.0"],
                    },
                    "section": {
                        "type": "string",
                        "title": "Section",
                        "description": "The "
                        "section, "
                        "form, "
                        "survey "
                        "instrument, "
                        "set "
                        "of "
                        "measures  "
                        "or "
                        "other "
                        "broad "
                        "category "
                        "used \n"
                        "to "
                        "group "
                        "variables. "
                        "Previously "
                        "called "
                        '"module."\n',
                        "examples": ["Demographics", "PROMIS", "Medical " "History"],
                    },
                    "name": {
                        "type": "string",
                        "title": "Variable " "Name",
                        "description": "The "
                        "name "
                        "of "
                        "a "
                        "variable "
                        "(i.e., "
                        "field) "
                        "as "
                        "it "
                        "appears "
                        "in "
                        "the "
                        "data. \n",
                        "examples": ["gender_id"],
                    },
                    "title": {
                        "type": "string",
                        "title": "Variable " "Label " "(ie " "Title)",
                        "description": "The "
                        "human-readable "
                        "title "
                        "or "
                        "label "
                        "of "
                        "the "
                        "variable.\n",
                        "examples": ["Gender " "identity"],
                    },
                    "description": {
                        "type": "string",
                        "title": "Variable " "Description",
                        "description": "An "
                        "extended "
                        "description "
                        "of "
                        "the "
                        "variable. "
                        "This "
                        "could "
                        "be "
                        "the "
                        "definition "
                        "of "
                        "a "
                        "variable "
                        "or "
                        "the \n"
                        "question "
                        "text "
                        "(e.g., "
                        "if "
                        "a "
                        "survey). \n",
                        "examples": [
                            "The "
                            "participant's "
                            "age "
                            "at "
                            "the "
                            "time "
                            "of "
                            "study "
                            "enrollment",
                            "What "
                            "is "
                            "the "
                            "highest "
                            "grade "
                            "or "
                            "level "
                            "of "
                            "school "
                            "you "
                            "have "
                            "completed "
                            "or "
                            "the "
                            "highest "
                            "degree "
                            "you "
                            "have "
                            "received?",
                        ],
                    },
                    "type": {
                        "title": "Variable " "Type",
                        "type": "string",
                        "description": "A "
                        "classification "
                        "or "
                        "category "
                        "of "
                        "a "
                        "particular "
                        "data "
                        "element "
                        "or "
                        "property "
                        "expected "
                        "or "
                        "allowed "
                        "in "
                        "the "
                        "dataset.\n",
                        "additionalDescription": "enum "
                        "definitions:\n"
                        "\n"
                        "-  "
                        "`number` "
                        "(A "
                        "numeric "
                        "value "
                        "with "
                        "optional "
                        "decimal "
                        "places. "
                        "(e.g., "
                        "3.14))\n"
                        "-  "
                        "`integer` "
                        "(A "
                        "whole "
                        "number "
                        "without "
                        "decimal "
                        "places. "
                        "(e.g., "
                        "42))\n"
                        "-  "
                        "`string` "
                        "(A "
                        "sequence "
                        "of "
                        "characters. "
                        "(e.g., "
                        '\\"test\\"))\n'
                        "-  "
                        "`any` "
                        "(Any "
                        "type "
                        "of "
                        "data "
                        "is "
                        "allowed. "
                        "(e.g., "
                        "true))\n"
                        "-  "
                        "`boolean` "
                        "(A "
                        "binary "
                        "value "
                        "representing "
                        "true "
                        "or "
                        "false. "
                        "(e.g., "
                        "true))\n"
                        "-  "
                        "`date` "
                        "(A "
                        "specific "
                        "calendar "
                        "date. "
                        "(e.g., "
                        '\\"2023-05-25\\"))\n'
                        "-  "
                        "`datetime` "
                        "(A "
                        "specific "
                        "date "
                        "and "
                        "time, "
                        "including "
                        "timezone "
                        "information. "
                        "(e.g., "
                        '\\"2023-05-25T10:30:00Z\\"))\n'
                        "-  "
                        "`time` "
                        "(A "
                        "specific "
                        "time "
                        "of "
                        "day. "
                        "(e.g., "
                        '\\"10:30:00\\"))\n'
                        "-  "
                        "`year` "
                        "(A "
                        "specific "
                        "year. "
                        "(e.g., "
                        "2023)\n"
                        "-  "
                        "`yearmonth` "
                        "(A "
                        "specific "
                        "year "
                        "and "
                        "month. "
                        "(e.g., "
                        '\\"2023-05\\"))\n'
                        "-  "
                        "`duration` "
                        "(A "
                        "length "
                        "of "
                        "time. "
                        "(e.g., "
                        '\\"PT1H\\")\n'
                        "-  "
                        "`geopoint` "
                        "(A "
                        "pair "
                        "of "
                        "latitude "
                        "and "
                        "longitude "
                        "coordinates. "
                        "(e.g., "
                        "[51.5074, "
                        "-0.1278]))\n",
                        "enum": [
                            "number",
                            "integer",
                            "string",
                            "any",
                            "boolean",
                            "date",
                            "datetime",
                            "time",
                            "year",
                            "yearmonth",
                            "duration",
                            "geopoint",
                        ],
                    },
                    "format": {
                        "title": "Variable " "Format",
                        "type": "string",
                        "description": "Indicates "
                        "the "
                        "format "
                        "of "
                        "the "
                        "type "
                        "specified "
                        "in "
                        "the "
                        "`type` "
                        "property. \n"
                        "Each "
                        "format "
                        "is "
                        "dependent "
                        "on "
                        "the "
                        "`type` "
                        "specified. \n"
                        "See "
                        "[here](https://specs.frictionlessdata.io/table-schema/#types-and-formats) \n"
                        "for "
                        "more "
                        "information "
                        "about "
                        "appropriate "
                        "`format` "
                        "values "
                        "by "
                        "variable "
                        "`type`.\n",
                        "additionalDescription": "examples/definitions "
                        "of "
                        "patterns "
                        "and "
                        "possible "
                        "values:\n"
                        "\n"
                        "Examples "
                        "of "
                        "date "
                        "time "
                        "pattern "
                        "formats\n"
                        "\n"
                        "- "
                        "`%Y-%m-%d` "
                        "(for "
                        "date, "
                        "e.g., "
                        "2023-05-25)\n"
                        "- "
                        "`%Y%-%d` "
                        "(for "
                        "date, "
                        "e.g., "
                        "20230525) "
                        "for "
                        "date "
                        "without "
                        "dashes\n"
                        "- "
                        "`%Y-%m-%dT%H:%M:%S` "
                        "(for "
                        "datetime, "
                        "e.g., "
                        "2023-05-25T10:30:45)\n"
                        "- "
                        "`%Y-%m-%dT%H:%M:%SZ` "
                        "(for "
                        "datetime "
                        "with "
                        "UTC "
                        "timezone, "
                        "e.g., "
                        "2023-05-25T10:30:45Z)\n"
                        "- "
                        "`%Y-%m-%dT%H:%M:%S%z` "
                        "(for "
                        "datetime "
                        "with "
                        "timezone "
                        "offset, "
                        "e.g., "
                        "2023-05-25T10:30:45+0300)\n"
                        "- "
                        "`%Y-%m-%dT%H:%M` "
                        "(for "
                        "datetime "
                        "without "
                        "seconds, "
                        "e.g., "
                        "2023-05-25T10:30)\n"
                        "- "
                        "`%Y-%m-%dT%H` "
                        "(for "
                        "datetime "
                        "without "
                        "minutes "
                        "and "
                        "seconds, "
                        "e.g., "
                        "2023-05-25T10)\n"
                        "- "
                        "`%H:%M:%S` "
                        "(for "
                        "time, "
                        "e.g., "
                        "10:30:45)\n"
                        "- "
                        "`%H:%M:%SZ` "
                        "(for "
                        "time "
                        "with "
                        "UTC "
                        "timezone, "
                        "e.g., "
                        "10:30:45Z)\n"
                        "- "
                        "`%H:%M:%S%z` "
                        "(for "
                        "time "
                        "with "
                        "timezone "
                        "offset, "
                        "e.g., "
                        "10:30:45+0300)\n"
                        "\n"
                        "Examples "
                        "of "
                        "string "
                        "formats\n"
                        "\n"
                        "- "
                        "`email` "
                        "if "
                        "valid "
                        "emails "
                        "(e.g., "
                        "test@gmail.com)\n"
                        "- "
                        "`uri` "
                        "if "
                        "valid "
                        "uri "
                        "addresses "
                        "(e.g., "
                        "https://example.com/resource123)\n"
                        "- "
                        "`binary` "
                        "if "
                        "a "
                        "base64 "
                        "binary "
                        "encoded "
                        "string "
                        "(e.g., "
                        "authentication "
                        "token "
                        "like "
                        "aGVsbG8gd29ybGQ=)\n"
                        "- "
                        "`uuid` "
                        "if "
                        "a "
                        "universal "
                        "unique "
                        "identifier "
                        "also "
                        "known "
                        "as "
                        "a "
                        "guid "
                        "(eg., "
                        "f47ac10b-58cc-4372-a567-0e02b2c3d479)\n"
                        "\n"
                        "\n"
                        "Examples "
                        "of "
                        "geopoint "
                        "formats\n"
                        "\n"
                        "The "
                        "two "
                        "types "
                        "of "
                        "formats "
                        "for "
                        "`geopoint` "
                        "(describing "
                        "a "
                        "geographic "
                        "point).\n"
                        "\n"
                        "- "
                        "`array` "
                        "(if "
                        "'lat,long' "
                        "(e.g., "
                        "36.63,-90.20))\n"
                        "- "
                        "`object` "
                        "(if "
                        "{'lat':36.63,'lon':-90.20})\n",
                    },
                    "constraints": {
                        "type": "object",
                        "properties": {
                            "required": {
                                "type": "boolean",
                                "title": "Required " "variable",
                                "description": "If "
                                "this "
                                "variable "
                                "is "
                                "marked "
                                "as "
                                "true, "
                                "then "
                                "this "
                                "variable's "
                                "value "
                                "must "
                                "be "
                                "present\n"
                                "(ie "
                                "not "
                                "missing; "
                                "see "
                                "missingValues). "
                                "If "
                                "marked "
                                "as "
                                "false "
                                "or "
                                "not "
                                "present, "
                                "then "
                                "the \n"
                                "variable "
                                "CAN "
                                "be "
                                "missing.\n",
                            },
                            "maxLength": {
                                "type": "integer",
                                "title": "Maximum " "Length",
                                "description": "Indicates "
                                "the "
                                "maximum "
                                "length "
                                "of "
                                "an "
                                "iterable "
                                "(e.g., "
                                "array, "
                                "string, "
                                "or\n"
                                "object). "
                                "For "
                                "example, "
                                "if "
                                "'Hello "
                                "World' "
                                "is "
                                "the "
                                "longest "
                                "value "
                                "of "
                                "a\n"
                                "categorical "
                                "variable, "
                                "this "
                                "would "
                                "be "
                                "a "
                                "maxLength "
                                "of "
                                "11.\n",
                            },
                            "enum": {
                                "type": "array",
                                "title": "Variable " "Possible " "Values",
                                "description": "Constrains "
                                "possible "
                                "values "
                                "to "
                                "a "
                                "set "
                                "of "
                                "values.\n",
                                "examples": [
                                    [1, 2, 3, 4, 5],
                                    [
                                        "Poor",
                                        "Fair",
                                        "Good",
                                        "Very " "good",
                                        "Excellent",
                                    ],
                                ],
                            },
                            "pattern": {
                                "type": "string",
                                "title": "Regular " "Expression " "Pattern",
                                "description": "A "
                                "regular "
                                "expression "
                                "pattern "
                                "the "
                                "data "
                                "MUST "
                                "conform "
                                "to.\n",
                            },
                            "maximum": {
                                "type": "integer",
                                "title": "Maximum " "Value",
                                "description": "Specifies "
                                "the "
                                "maximum "
                                "value "
                                "of "
                                "a "
                                "field "
                                "(e.g., "
                                "maximum "
                                "-- "
                                "or "
                                "most\n"
                                "recent "
                                "-- "
                                "date, "
                                "maximum "
                                "integer "
                                "etc). "
                                "Note, "
                                "this "
                                "is "
                                "different "
                                "then\n"
                                "maxLength "
                                "property.\n",
                            },
                            "minimum": {
                                "type": "integer",
                                "title": "Minimum " "Value",
                                "description": "Specifies "
                                "the "
                                "minimum "
                                "value "
                                "of "
                                "a "
                                "field.\n",
                            },
                        },
                    },
                    "enumLabels": {
                        "title": "Variable "
                        "Value "
                        "Encodings "
                        "(i.e., "
                        "mappings; "
                        "value "
                        "labels)",
                        "description": "Variable "
                        "value "
                        "encodings "
                        "provide "
                        "a "
                        "way "
                        "to "
                        "further "
                        "annotate "
                        "any "
                        "value "
                        "within "
                        "a "
                        "any "
                        "variable "
                        "type,\n"
                        "making "
                        "values "
                        "easier "
                        "to "
                        "understand. \n"
                        "\n"
                        "\n"
                        "Many "
                        "analytic "
                        "software "
                        "programs "
                        "(e.g., "
                        "SPSS,Stata, "
                        "and "
                        "SAS) "
                        "use "
                        "numerical "
                        "encodings "
                        "and "
                        "some "
                        "algorithms\n"
                        "only "
                        "support "
                        "numerical "
                        "values. "
                        "Encodings "
                        "(and "
                        "mappings) "
                        "allow "
                        "categorical "
                        "values "
                        "to "
                        "be "
                        "stored "
                        "as\n"
                        "numerical "
                        "values.\n"
                        "\n"
                        "Additionally, "
                        "as "
                        "another "
                        "use "
                        "case, "
                        "this "
                        "field "
                        "provides "
                        "a "
                        "way "
                        "to\n"
                        "store "
                        "categoricals "
                        "that "
                        "are "
                        "stored "
                        "as  "
                        '"short" '
                        "labels "
                        "(such "
                        "as\n"
                        "abbreviations).\n"
                        "\n"
                        "This "
                        "field "
                        "is "
                        "intended "
                        "to "
                        "follow "
                        "[this "
                        "pattern](https://specs.frictionlessdata.io/patterns/#table-schema-enum-labels-and-ordering)\n",
                        "type": "object",
                        "examples": [
                            {
                                "1": "Poor",
                                "2": "Fair",
                                "3": "Good",
                                "4": "Very " "good",
                                "5": "Excellent",
                            },
                            {
                                "HW": "Hello " "world",
                                "GBW": "Good " "bye " "world",
                                "HM": "Hi, " "Mike",
                            },
                        ],
                    },
                    "enumOrdered": {
                        "title": "An " "ordered " "variable",
                        "description": "Indicates "
                        "whether "
                        "a "
                        "categorical "
                        "variable "
                        "is "
                        "ordered. "
                        "This "
                        "variable  "
                        "is\n"
                        "relevant "
                        "for "
                        "variables "
                        "that "
                        "have "
                        "an "
                        "ordered "
                        "relationship "
                        "but "
                        "not\n"
                        "necessarily  "
                        "a "
                        "numerical "
                        "relationship "
                        "(e.g., "
                        "Strongly "
                        "disagree "
                        "< "
                        "Disagree\n"
                        "< "
                        "Neutral "
                        "< "
                        "Agree).\n"
                        "\n"
                        "This "
                        "field "
                        "is "
                        "intended "
                        "to "
                        "follow "
                        "the "
                        "ordering "
                        "aspect "
                        "of "
                        "this "
                        "[this "
                        "pattern][this "
                        "pattern](https://specs.frictionlessdata.io/patterns/#table-schema-enum-labels-and-ordering)\n",
                        "type": "boolean",
                    },
                    "missingValues": {
                        "title": "Missing " "Values",
                        "description": "A "
                        "list "
                        "of "
                        "missing "
                        "values "
                        "specific "
                        "to "
                        "a "
                        "variable.\n",
                        "examples": [
                            ["Missing", "Skipped", "No " "preference"],
                            ["Missing"],
                        ],
                        "type": "array",
                    },
                    "trueValues": {
                        "title": "Boolean " "True " "Value " "Labels",
                        "description": "For "
                        "boolean "
                        "(true) "
                        "variable "
                        "(as "
                        "defined "
                        "in "
                        "type "
                        "field), "
                        "this "
                        "field "
                        "allows\n"
                        "a "
                        "physical "
                        "string "
                        "representation "
                        "to "
                        "be "
                        "cast "
                        "as "
                        "true "
                        "(increasing\n"
                        "readability "
                        "of "
                        "the "
                        "field). "
                        "It "
                        "can "
                        "include "
                        "one "
                        "or "
                        "more "
                        "values.\n",
                        "type": "array",
                        "examples": [["required", "Yes", "Checked"], ["required"]],
                    },
                    "falseValues": {
                        "title": "Boolean " "False " "Value " "Labels",
                        "description": "For "
                        "boolean "
                        "(false) "
                        "variable "
                        "(as "
                        "defined "
                        "in "
                        "type "
                        "field), "
                        "this "
                        "field "
                        "allows\n"
                        "a "
                        "physical "
                        "string "
                        "representation "
                        "to "
                        "be "
                        "cast "
                        "as "
                        "false "
                        "(increasing\n"
                        "readability "
                        "of "
                        "the "
                        "field) "
                        "that "
                        "is "
                        "not "
                        "a "
                        "standard "
                        "false "
                        "value. "
                        "It "
                        "can "
                        "include "
                        "one "
                        "or "
                        "more "
                        "values.\n",
                        "type": "array",
                        "examples": [["Not " "required", "NOT " "REQUIRED"], ["No"]],
                    },
                    "custom": {
                        "type": "object",
                        "description": "Additional "
                        "properties "
                        "not "
                        "included "
                        "a "
                        "core "
                        "property. \n",
                    },
                    "standardsMappings": {
                        "type": "array",
                        "description": "\n"
                        "A "
                        "set "
                        "of "
                        "instrument "
                        "and "
                        "item "
                        "references "
                        "to "
                        "standardized "
                        "data "
                        "elements "
                        "designed "
                        "to "
                        "document\n"
                        "the "
                        "[HEAL "
                        "common "
                        "data "
                        "elements "
                        "program](https://heal.nih.gov/data/common-data-elements)\n"
                        "and "
                        "other "
                        "standardized/common "
                        "element "
                        "sources "
                        "to "
                        "facilitate "
                        "cross-study "
                        "comparison "
                        "and "
                        "interoperability\n"
                        "of "
                        "data. "
                        "One "
                        "can "
                        "either "
                        "map "
                        "an "
                        "individual "
                        "data "
                        "element "
                        "or "
                        "an "
                        "instrument "
                        "in "
                        "which "
                        "the "
                        "field "
                        "is \n"
                        "a "
                        "part "
                        "of.\n"
                        "\n"
                        "__**All "
                        "Fields "
                        "Mapped "
                        "(Both "
                        "Instrument "
                        "and "
                        "Item)**__\n"
                        "\n"
                        "```json\n"
                        '"standardsMappings": '
                        "[\n"
                        "    "
                        "{\n"
                        "        "
                        '"instrument": '
                        "{\n"
                        "            "
                        '"url": '
                        '"https://www.heal.nih.gov/files/CDEs/2023-05/adult-demographics-cdes.xlsx",\n'
                        "            "
                        '"source": '
                        '"heal-cde",\n'
                        "            "
                        '"title": '
                        '"adult-demographics",\n'
                        "            "
                        '"id": '
                        '"5141"\n'
                        "        "
                        "},\n"
                        "        "
                        '"item": '
                        "{\n"
                        "            "
                        '"url": '
                        '"https://evs.nci.nih.gov/ftp1/CDISC/SDTM/SDTM%20Terminology.html#CL.C74457.RACE",\n'
                        "            "
                        '"source": '
                        '"CDISC",\n'
                        "            "
                        '"id": '
                        '"C74457"\n'
                        "        "
                        "}\n"
                        "    "
                        "}\n"
                        "]\n"
                        "```\n"
                        "\n"
                        "__**Only "
                        "Instrument "
                        "Title "
                        "of "
                        "Form "
                        "CDE "
                        "File "
                        "Mapped**__\n"
                        "\n"
                        "In "
                        "this "
                        "scenario, "
                        "especially "
                        "as "
                        "CDE "
                        "variables "
                        "do "
                        "not "
                        "have "
                        "associated "
                        "CDISC "
                        "ids "
                        "listed, "
                        "only "
                        "instrument "
                        "information "
                        "is "
                        "given.\n"
                        "\n"
                        "```json\n"
                        '"standardsMappings": '
                        "[\n"
                        "    "
                        "{\n"
                        "        "
                        '"instrument": '
                        "{\n"
                        "            "
                        '"source": '
                        '"heal-cde",\n'
                        "            "
                        '"title": '
                        '"Adult '
                        'demographics"\n'
                        "        "
                        "}\n"
                        "    "
                        "}\n"
                        "]\n"
                        "```\n"
                        "\n"
                        "__**Only "
                        "Instrument "
                        "ID "
                        "of "
                        "HEAL "
                        "CDE "
                        "Mapped**__\n"
                        "\n"
                        "```json\n"
                        '"standardsMappings": '
                        "[\n"
                        "    "
                        "{\n"
                        "        "
                        '"instrument": '
                        "{\n"
                        "            "
                        '"source": '
                        '"heal-cde",\n'
                        "            "
                        '"id": '
                        '"5141"\n'
                        "        "
                        "}\n"
                        "    "
                        "}\n"
                        "]\n"
                        "```\n"
                        "\n"
                        "__**Other "
                        "Non-HEAL "
                        "CDE "
                        "Use "
                        "Cases**__\n"
                        "\n"
                        "Only "
                        "item "
                        "matched "
                        "(for "
                        "example "
                        "if "
                        "found "
                        "in "
                        "the "
                        "NIH "
                        "(not "
                        "HEAL) "
                        "CDE "
                        "repository). "
                        "Folks "
                        "would "
                        "enter "
                        "the "
                        "information "
                        "in "
                        "the "
                        '"Identifier" '
                        "section. "
                        "Similar "
                        "to "
                        "the "
                        "above, "
                        "they "
                        "could "
                        "also "
                        "just "
                        "enter "
                        "the "
                        '"url".\n'
                        "\n"
                        "```json\n"
                        '"standardsMappings": '
                        "[\n"
                        "    "
                        "{\n"
                        "        "
                        '"item": '
                        "{\n"
                        "            "
                        '"source": '
                        '"NLM",\n'
                        "            "
                        '"id": '
                        '"Fakc6Jy2x"\n'
                        "        "
                        "}\n"
                        "    "
                        "}\n"
                        "]\n"
                        "```\n"
                        "\n"
                        "__**Multiple "
                        "CDE "
                        "Mappings**__\n"
                        "\n"
                        "Two "
                        "separate "
                        "records. "
                        "If "
                        "desired, "
                        "multiple "
                        "standard "
                        "mappings "
                        "can "
                        "be "
                        "entered, "
                        "say "
                        "from "
                        "the "
                        "NIH "
                        "HEAL "
                        "CDE "
                        "repo "
                        "and "
                        "the "
                        "NIH "
                        "CDE "
                        "lookup "
                        "(NLM) "
                        "by "
                        "way "
                        "of "
                        "two "
                        "separate "
                        "records "
                        "in "
                        "the "
                        "list.\n"
                        "\n"
                        "```json\n"
                        '"standardsMappings": '
                        "[\n"
                        "    "
                        "{\n"
                        "        "
                        '"instrument": '
                        "{\n"
                        "            "
                        '"source": '
                        '"heal-cde",\n'
                        "            "
                        '"title": '
                        '"Adult '
                        'demographics"\n'
                        "        "
                        "},\n"
                        "        "
                        '"item": '
                        "{\n"
                        "            "
                        '"source": '
                        '"CDISC",\n'
                        "            "
                        '"id": '
                        '"C74457"\n'
                        "        "
                        "},\n"
                        "    "
                        "},\n"
                        "    "
                        "{\n"
                        "        "
                        '"item": '
                        "{\n"
                        "            "
                        '"source": '
                        '"NLM",\n'
                        "            "
                        '"id": '
                        '"Fakc6Jy2x"\n'
                        "        "
                        "}\n"
                        "    "
                        "}\n"
                        "]\n"
                        "```\n",
                        "items": {
                            "type": "object",
                            "properties": {
                                "instrument": {
                                    "type": "object",
                                    "title": "Standard " "mapping " "- " "instrument",
                                    "description": "A "
                                    "standardized "
                                    "set "
                                    "of "
                                    "items "
                                    "which "
                                    "encompass \n"
                                    "a "
                                    "variable "
                                    "in "
                                    "this "
                                    "variable "
                                    "level "
                                    "metadata "
                                    "document "
                                    "(if "
                                    "at "
                                    "the "
                                    "root "
                                    "level "
                                    "or "
                                    "the "
                                    "document "
                                    "level) \n"
                                    "or "
                                    "the "
                                    "individual "
                                    "variable "
                                    "(if "
                                    "at "
                                    "the "
                                    "field "
                                    "level). \n"
                                    "\n"
                                    "\n"
                                    "!!! "
                                    "note "
                                    '"NOTE"\n'
                                    "\n"
                                    "  "
                                    "If "
                                    "information "
                                    "is "
                                    "present "
                                    "at "
                                    "both "
                                    "the "
                                    "root "
                                    "and "
                                    "the "
                                    "field "
                                    "level, \n"
                                    "  "
                                    "then "
                                    "the "
                                    "information "
                                    "at "
                                    "the "
                                    "field "
                                    "level "
                                    "would "
                                    "take "
                                    "precedence "
                                    "(i.e., "
                                    "it "
                                    "would "
                                    "cascade).\n",
                                    "properties": {
                                        "url": {
                                            "title": "Url",
                                            "description": "A "
                                            "url "
                                            "(e.g., "
                                            "link, "
                                            "address) "
                                            "to "
                                            "a "
                                            "file "
                                            "or "
                                            "other "
                                            "resource "
                                            "containing "
                                            "the "
                                            "instrument, "
                                            "or\n"
                                            "a "
                                            "set "
                                            "of "
                                            "items "
                                            "which "
                                            "encompass "
                                            "a "
                                            "variable "
                                            "in "
                                            "this "
                                            "variable "
                                            "level "
                                            "metadata "
                                            "document "
                                            "(if "
                                            "at "
                                            "the "
                                            "root "
                                            "level "
                                            "or "
                                            "the "
                                            "document "
                                            "level) \n"
                                            "or "
                                            "the "
                                            "individual "
                                            "variable "
                                            "(if "
                                            "at "
                                            "the "
                                            "field "
                                            "level). \n",
                                            "type": "string",
                                            "format": "uri",
                                            "examples": [
                                                "https://www.heal.nih.gov/files/CDEs/2023-05/adult-demographics-cdes.xlsx"
                                            ],
                                        },
                                        "source": {
                                            "type": "string",
                                            "title": "Source",
                                            "description": "An "
                                            "abbreviated "
                                            "name/acronym "
                                            "from "
                                            "a "
                                            "controlled "
                                            "vocabulary "
                                            "referencing "
                                            "the "
                                            "resource "
                                            "(e.g., "
                                            "program "
                                            "or "
                                            "repository)\n"
                                            "containing "
                                            "the "
                                            "instrument, "
                                            "or "
                                            "a "
                                            "set "
                                            "of "
                                            "items "
                                            "which "
                                            "encompass "
                                            "a "
                                            "variable "
                                            "in "
                                            "this "
                                            "variable "
                                            "level "
                                            "metadata "
                                            "document "
                                            "(if "
                                            "at "
                                            "the "
                                            "root "
                                            "level "
                                            "or "
                                            "the "
                                            "document "
                                            "level) \n"
                                            "or "
                                            "the "
                                            "individual "
                                            "variable "
                                            "(if "
                                            "at "
                                            "the "
                                            "field "
                                            "level). \n",
                                            "enum": ["heal-cde"],
                                        },
                                        "title": {
                                            "type": "string",
                                            "title": "Title",
                                            "examples": [
                                                "Adult " "demographics",
                                                "adult-demographics",
                                            ],
                                        },
                                        "id": {
                                            "type": "string",
                                            "title": "Identifier",
                                            "description": "A "
                                            "code "
                                            "or "
                                            "other "
                                            "string "
                                            "that "
                                            "identifies "
                                            "the "
                                            "instrument "
                                            "within "
                                            "the "
                                            "source.\n"
                                            "This "
                                            "should "
                                            "always "
                                            "be "
                                            "from "
                                            "the "
                                            "source's "
                                            "formal, "
                                            "standardized "
                                            "identification "
                                            "system \n",
                                            "examples": ["5141"],
                                        },
                                    },
                                },
                                "item": {
                                    "type": "object",
                                    "title": "Standards " "mappings " "- " "Item",
                                    "description": "A "
                                    "standardized "
                                    "item "
                                    "(ie "
                                    "field, "
                                    "variable "
                                    "etc) "
                                    "mapped "
                                    "to "
                                    "this "
                                    "individual "
                                    "variable.\n",
                                    "properties": {
                                        "url": {
                                            "title": "Standards "
                                            "mappings "
                                            "- "
                                            "Url",
                                            "description": "The "
                                            "url "
                                            "that "
                                            "links "
                                            "out "
                                            "to "
                                            "the "
                                            "published, "
                                            "standardized "
                                            "mapping "
                                            "of "
                                            "a "
                                            "variable "
                                            "(e.g., "
                                            "common "
                                            "data "
                                            "element)\n",
                                            "type": "string",
                                            "format": "uri",
                                            "examples": [
                                                "https://evs.nci.nih.gov/ftp1/CDISC/SDTM/SDTM%20Terminology.html#CL.C74457.RACE"
                                            ],
                                        },
                                        "source": {
                                            "title": "Standards "
                                            "mappings "
                                            "- "
                                            "Source",
                                            "description": "The "
                                            "source "
                                            "of "
                                            "the "
                                            "standardized "
                                            "variable. "
                                            "Note, "
                                            "this "
                                            "property "
                                            "is "
                                            "required "
                                            "if \n"
                                            "an "
                                            "id "
                                            "is "
                                            "specified.\n",
                                            "examples": ["CDISC"],
                                            "type": "string",
                                        },
                                        "id": {
                                            "title": "Standards " "Mappings " "- " "Id",
                                            "type": "string",
                                            "description": "The "
                                            "id "
                                            "locating "
                                            "the "
                                            "individual "
                                            "mapping "
                                            "within "
                                            "the "
                                            "given "
                                            "source. \n"
                                            "Note, "
                                            "the "
                                            "`standardsMappings[0].source` "
                                            "property "
                                            "is "
                                            "required "
                                            "if \n"
                                            "this "
                                            "property "
                                            "is "
                                            "specified.\n",
                                            "examples": ["C74457"],
                                        },
                                    },
                                },
                            },
                        },
                    },
                    "relatedConcepts": {
                        "title": "Related " "Concepts",
                        "description": "__**[Under "
                        "development]**__ "
                        "Mappings "
                        "to "
                        "a "
                        "published "
                        "set "
                        "of "
                        "concepts "
                        "related "
                        "to "
                        "the "
                        "given "
                        "field "
                        "such "
                        "as \n"
                        "ontological "
                        "information "
                        "(eg., "
                        "NCI "
                        "thesaurus, "
                        "bioportal "
                        "etc)\n",
                        "type": "array",
                        "items": {
                            "type": "object",
                            "properties": {
                                "url": {
                                    "title": "Related " "Concepts " "- " "Url",
                                    "description": "The "
                                    "url "
                                    "that "
                                    "links "
                                    "out "
                                    "to "
                                    "the "
                                    "published, "
                                    "related "
                                    "concept. \n"
                                    "The "
                                    "listed "
                                    "examples "
                                    "could "
                                    "both "
                                    "be "
                                    "attached "
                                    "to "
                                    "any "
                                    "variable "
                                    "related "
                                    "to, "
                                    "for "
                                    "example, "
                                    "heroin "
                                    "use.\n",
                                    "type": "string",
                                    "format": "uri",
                                    "examples": [
                                        "https://www.ebi.ac.uk/chebi/chebiOntology.do?chebiId=CHEBI:27808",
                                        "http://purl.bioontology.org/ontology/RXNORM/3304",
                                    ],
                                },
                                "title": {
                                    "title": "Related " "concepts " "- " "Type",
                                    "description": "A "
                                    "human-readable "
                                    "title "
                                    "(ie "
                                    "label) "
                                    "to "
                                    "a "
                                    "concept "
                                    "related "
                                    "to "
                                    "the "
                                    "given "
                                    "field.\n"
                                    "The "
                                    "listed "
                                    "examples "
                                    "could "
                                    "both "
                                    "be "
                                    "attached "
                                    "to "
                                    "any "
                                    "variable "
                                    "related "
                                    "to, "
                                    "for "
                                    "example, "
                                    "heroin "
                                    "use.\n",
                                    "type": "string",
                                    "examples": [
                                        "Heroin " "Molecular " "Structure",
                                        "Heroin " "Ontology",
                                    ],
                                },
                                "source": {
                                    "title": "Related " "Concepts " "- " "Source",
                                    "description": "The "
                                    "source "
                                    "(e.g., "
                                    "a "
                                    "dictionary "
                                    "or "
                                    "vocabulary "
                                    "set) "
                                    "to "
                                    "a "
                                    "concept "
                                    "related "
                                    "to "
                                    "the "
                                    "given "
                                    "field.\n"
                                    "The "
                                    "listed "
                                    "examples "
                                    "could "
                                    "both "
                                    "be "
                                    "attached "
                                    "to "
                                    "any "
                                    "variable "
                                    "related "
                                    "to, "
                                    "for "
                                    "example, "
                                    "heroin "
                                    "use.\n",
                                    "type": "string",
                                    "examples": ["CHEBI", "RXNORM"],
                                },
                                "id": {
                                    "title": "Related " "Concepts " "- " "Id",
                                    "type": "string",
                                    "description": "The "
                                    "id "
                                    "locating "
                                    "the "
                                    "individual "
                                    "concept "
                                    "within "
                                    "the "
                                    "source "
                                    "of "
                                    "the "
                                    "given "
                                    "field.\n"
                                    "The "
                                    "listed "
                                    "examples "
                                    "could "
                                    "both "
                                    "be "
                                    "attached "
                                    "to "
                                    "any "
                                    "variable "
                                    "related "
                                    "to, "
                                    "for "
                                    "example, "
                                    "heroin "
                                    "use.\n",
                                    "examples": ["27808", "3304"],
                                },
                            },
                        },
                    },
                },
            },
        },
    },
    "propertyNames": {
        "description": "To allow additional properties for "
        "compatibility with other standards at the "
        '"table" , or root, but not included in '
        "the core `properties` set:\n"
        "\n"
        "[Frictionless Data package table schema "
        "standard](https://specs.frictionlessdata.io/table-schema): "
        "`missingValues`|`primaryKey`|`foreignKeys`\n",
        "enum": [
            "title",
            "description",
            "schemaVersion",
            "version",
            "standardsMappings",
            "fields",
            "custom",
            "missingValues",
            "primaryKey",
            "foreignKeys",
        ],
    },
}
